import sys
import os
sys.path.insert(0, os.path.abspath('..'))
import gents
# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'GenTS'
copyright = '2025, Cameron Cummins'
author = 'Cameron Cummins'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary'
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '.ipynb_checkpoints']


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'alabaster'
html_static_path = ['_static']



from netCDF4 import Dataset
from pathlib import Path
from gents.meta import get_attributes, get_time_variables_names
from cftime import num2date
import numpy as np


def get_concat_coords(hf_datasets):
    dim_coords = {}

    for ds in hf_datasets:
        for dim in ds.dimensions:
            if dim in ds.variables and dim in dim_coords:
                dim_coords[dim] = np.unique(np.concat([ds[dim][:], dim_coords[dim]]))
            elif dim in ds.variables:
                dim_coords[dim] = ds[dim][:]
            else:
                dim_coords[dim] = np.arange(ds.dimensions[dim].size)

    return dim_coords


def get_timestamp_str(times):
    """
    Creates timestamp string to describe time range for netCDF dataset

    :param times: Time values for netCDF dataset in integer form with units and calendar attributes.
    :return: String containing appropriate timestamp.
    """
    calendar = times.calendar
    units = times.units
    data = np.sort(times[:])
    
    start_t = num2date(data[0], units=units, calendar=calendar)
    if times.shape[0] == 1:
        return start_t.strftime("%Y%m%d%H")
    else:
        dt = num2date(data[1], units=units, calendar=calendar) - start_t
        minutes = dt.total_seconds() / 60
        hours = minutes / 60
        days = hours / 24
        months = days / 30
    
        if minutes < 1:
            time_format = "%Y%m%d%H%M%S"
        elif hours < 1:
            time_format = "%Y%m%d%H"
        elif days < 1:
            time_format = "%Y%m%d"
        elif months < 1:
            time_format = "%Y%m"
        else:
            time_format = "%Y"
        
        end_t = num2date(data[-1], units=units, calendar=calendar)
        return f"{start_t.strftime(time_format)}-{end_t.strftime(time_format)}"


class MHFDataset:
    def __init__(self, hf_paths):
        self.__hf_files = [Path(path) for path in hf_paths]
        self.__hf_datasets = None
        self.__time_mapping = {}
        self.__data_coords = None

        with Dataset(self.__hf_files[0], 'r') as ds:
            self.__time_name, self.time_bnds_name = get_time_variables_names(ds)

    def open(self):
        if self.__hf_datasets is None:
            self.__hf_datasets = [Dataset(path, 'r') for path in self.__hf_files]
            for hf_index in range(len(self.__hf_datasets)):
                time_vals = np.squeeze(self.__hf_datasets[hf_index][self.__time_name][:])
                if len(time_vals.shape) == 0:
                    time_vals = [float(time_vals)]
                
                
                for time in time_vals:
                    time = float(time)
                    if time in self.__time_mapping:
                        self.__time_mapping[time].append(hf_index)
                    else:
                        self.__time_mapping[time] = [hf_index]
            if not self.is_time_consistent():
                raise Exception("Fragmentation is not consistent over time.")

    def close(self):
        for ds in self.__hf_datasets:
            ds.close()

    def get_time_vals(self):
        return list(self.__time_mapping.keys())

    def get_timestamp_string(self):
        return get_timestamp_str(self.__hf_datasets[0][self.__time_name])

    def is_time_consistent(self):
        n_time_files = len(self.__time_mapping[self.get_time_vals()[0]])
        for time in self.__time_mapping:
            if len(self.__time_mapping[time]) != n_time_files:
                return False
        return True

    def is_fragmented(self):
        init_time = self.get_time_vals()[0]
        if len(self.__time_mapping[init_time]) > 1:
            return True
        return False

    def get_var_dimensions(self, var_name):
        init_ds = self.__hf_datasets[0][var_name]
        return list(init_ds.dimensions)

    def get_var_dtype(self, var_name):
        return self.__hf_datasets[0][var_name].dtype

    def get_var_attrs(self, var_name):
        return get_attributes(self.__hf_datasets[0][var_name])

    def __check_coord_map(self):
        if self.__data_coords is None:
            if self.is_fragmented():
                self.__data_coords = get_concat_coords(self.__hf_datasets)
            else:
                self.__data_coords = {}
                init_ds = self.__hf_datasets[0]
                for dim in init_ds.dimensions:
                    if dim in init_ds.variables:
                        self.__data_coords[dim] = init_ds[dim][:]
                    else:
                        self.__data_coords[dim] = np.arange(init_ds.dimensions[dim].size)
                self.__data_coords[self.__time_name] = self.get_time_vals()

    def get_var_data_shape(self, var_name):
        self.__check_coord_map()
        if var_name in self.__data_coords:
            return [len(self.__data_coords[var_name])]
        else:
            init_ds = self.__hf_datasets[0][var_name]
            dim_shape = []
            if self.__time_name in init_ds.dimensions:
                dim_shape.append(len(self.get_time_vals()))
            dim_shape += [len(self.__data_coords[dim]) for dim in init_ds.dimensions if dim != self.__time_name]
            return dim_shape

    def get_var_vals(self, var_name, time_index_start=0, time_index_end=None):
        self.__check_coord_map()
        if var_name in self.__data_coords:
            return self.__data_coords[var_name]

        if "time" not in self.get_var_dimensions(var_name):
            return self.__hf_datasets[0][var_name][:]

        time_vals = self.get_time_vals()[time_index_start:time_index_end]
        data_shape = self.get_var_data_shape(var_name)
        data_shape[0] = len(time_vals)

        var_vals = np.empty(data_shape, dtype=self.__hf_datasets[0][var_name].dtype)
        if not self.is_fragmented():
            for index, time_val in enumerate(time_vals):
                hf_data = self.__hf_datasets[self.__time_mapping[time_val][0]]
                if hf_data[self.__time_name].shape[0] > 1:
                    sub_t_index = np.where(hf_data[self.__time_name][:] == time_val)[0]
                    var_vals[index] = hf_data[var_name][:][sub_t_index]
                else:
                    var_vals[index] = hf_data[var_name][:]
        else:
            for time_index, time_val in enumerate(time_vals):
                for hf_index in self.__time_mapping[time_val]:
                    hf_data = self.__hf_datasets[hf_index]
                    if self.__time_name in hf_data[var_name].dimensions and hf_data[self.__time_name].shape[0] > 1:
                        sub_t_index = np.where(hf_data[self.__time_name][:] == time_val)[0]
                        hf_data_fragment = hf_data[var_name][:][sub_t_index]
                    else:
                        hf_data_fragment = hf_data[var_name][:]
                    
                    index_ranges = []
                    for dim_index, dim in enumerate(hf_data[var_name].dimensions):
                        if dim == self.__time_name:
                            index_ranges.append(time_index)
                        elif dim in hf_data.variables:
                            dim_vals = hf_data[dim][:]
                            lower_index = np.where(np.min(dim_vals) == self.__data_coords[dim])[0][0]
                            upper_index = np.where(np.max(dim_vals) == self.__data_coords[dim])[0][0]
                            if lower_index == upper_index:
                                index_ranges.append(lower_index)
                            else:
                                index_ranges.append(slice(lower_index, upper_index+1))
                        elif var_vals.shape[dim_index] == 1:
                            index_ranges.append(0)
                        else:
                            index_ranges.append(slice(0, var_vals.shape[dim_index]))
                    
                    var_vals[tuple(index_ranges)] = np.squeeze(hf_data_fragment)
        return var_vals

    def get_global_attrs(self):
        assert self.__hf_datasets is not None

        agg_attrs = {}
        for ds in self.__hf_datasets:
            agg_attrs |= get_attributes(ds)
        return agg_attrs

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc, tb):
        self.close()
        return False

    def __contains__(self, key):
        return key in self.__hf_files

    def __iter__(self):
        return iter(self.__hf_datasets)

    def __len__(self):
        return len(self.__hf_datasets)
    
    def __getitem__(self, item):
        return self.__hf_datasets[item]


#!/usr/bin/env python
"""
utils.py

Developer: Cameron Cummins
Contact: cameron.cummins@utexas.edu
Last Header Update: 07/03/25
"""
from time import time
from importlib.metadata import version
from netCDF4 import Dataset
import logging
import sys
import datetime
import numpy as np

LOG_LEVEL_IO_WARNING = 5


def get_time_stamp():
    """Returns system date-time timestamp."""
    return datetime.datetime.fromtimestamp(time()).strftime('%Y-%m-%d %H:%M')


def get_version():
    """Returns the version of the installed GenTS package."""
    return version('gents')


def enable_logging(verbose=False, output_path=None):
    """
    Initializes logger to output GenTS logs.
    
    :param verbose: If true, all logs will be output including messages for individual file paths. Otherwise, only messages with level DEBUG or higher will be output. (Defaults to False)
    :param output_path: File path to output logs to. (Defaults to None)
    """
    logger = logging.getLogger("gents")

    formatter = logging.Formatter("%(asctime)s [%(levelname)s] %(name)s: %(message)s")
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setFormatter(formatter)

    if output_path:
        file_handler = logging.FileHandler(output_path)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    if verbose:
        logger.setLevel(LOG_LEVEL_IO_WARNING)
    else:
        logger.setLevel(logging.DEBUG)

    logger.addHandler(stdout_handler)

    logger.info(f"GenTS version {get_version()}")
    logger.info(f"Logging enabled (verbose={verbose}, output_path={output_path})")


class ProgressBar:
    def __init__(self, total, length=40):
        """
        Progress bar for visualizing various processes throughout the package.

        :param total: Total number of iterations in the loop.
        :param length: Length of the progress bar in characters. (Default is 40)
        """
        self.total = total
        self.length = length
        self.start_time = time()
        self.count = 0

    def step(self):
        """Update the progress bar by a given step."""
        self.count += 1
        percent = self.count / self.total
        filled_length = int(self.length * percent)
        bar = "█" * filled_length + "-" * (self.length - filled_length)
        elapsed = time() - self.start_time
        sys.stdout.write(
            f"\r|{bar}| {percent:6.2%}  {self.count}/{self.total}  Elapsed: {elapsed:5.1f}s"
        )
        sys.stdout.flush()

        if self.count >= self.total:
            sys.stdout.write("\n")


from os import listdir
from netCDF4 import Dataset
from cftime import num2date
import numpy as np
from gents.tests.test_cases import *


def test_simple_case(simple_case):
    input_head_dir, output_head_dir = simple_case
    assert len(listdir(input_head_dir)) == SIMPLE_NUM_TEST_HIST_FILES
    times = []
    time_bounds = []
    calendar = None
    units = None
    for file_name in listdir(input_head_dir):
        with Dataset(f"{input_head_dir}/{file_name}", 'r') as hf_ds:
            assert "time" in hf_ds.variables
            assert "units" in hf_ds["time"].ncattrs()
            units = hf_ds["time"].units

            assert "calendar" in hf_ds["time"].ncattrs()
            calendar = hf_ds["time"].calendar

            time = hf_ds["time"][:][0]
            bounds = hf_ds["time_bounds"][:][0]

            assert time >= 0
            assert time >= bounds[0]
            assert time <= bounds[1]
            assert bounds[0] != bounds[1]

            times.append(time)
            time_bounds.append(bounds)
    
    start_date = num2date(np.min(times), units=units, calendar=calendar)
    end_date = num2date(np.max(times), units=units, calendar=calendar)
    assert (end_date - start_date).days == (SIMPLE_NUM_TEST_HIST_FILES-1)*30
    assert start_date == num2date(0.5*30, units=units, calendar=calendar)


def test_unstructured_grid_case(unstructured_grid_case):
    input_head_dir, output_head_dir = unstructured_grid_case
    assert len(listdir(input_head_dir)) == SIMPLE_NUM_TEST_HIST_FILES

    with Dataset(f"{input_head_dir}/{listdir(input_head_dir)[0]}", 'r') as hf_ds:
        assert hf_ds["VAR_AUX_0"].size == UNSTRUCT_GRID_NUM_NCOLS
        assert hf_ds["VAR_AUX_0"].dimensions == ("ncol",)
        assert hf_ds["VAR0"].dimensions == ("time", "ncol",)


def test_time_bounds_case(time_bounds_case):
    input_head_dir, output_head_dir = time_bounds_case
    assert len(listdir(input_head_dir)) == TIME_NUM_TEST_HIST_FILES
    for file_name in listdir(input_head_dir):
        with Dataset(f"{input_head_dir}/{file_name}", 'r') as hf_ds:
            assert 'Time_Bounds' in hf_ds.variables
            assert 'Time' in hf_ds.variables


def test_scrambled_case(scrambled_case):
    input_head_dir, output_head_dir = scrambled_case
    assert len(listdir(input_head_dir)) == SCRAMBLED_NUM_TEST_HIST_FILES


def test_structured_case(structured_case):
    input_head_dir, output_head_dir = structured_case
    assert len(listdir(input_head_dir)) == STRUCTURED_NUM_DIRS
    for top_dir in listdir(input_head_dir):
        assert len(listdir(f"{input_head_dir}/{top_dir}")) == STRUCTURED_NUM_SUBDIRS
        for sub_dir in listdir(f"{input_head_dir}/{top_dir}/"):
            assert len(listdir(f"{input_head_dir}/{top_dir}/{sub_dir}")) == STRUCTURED_NUM_TEST_HIST_FILES


def test_time_bounds_missing_attrs_case(simple_case_missing_attrs):
    input_head_dir, output_head_dir = simple_case_missing_attrs
    for file_name in listdir(input_head_dir):
        with Dataset(f"{input_head_dir}/{file_name}", 'r') as hf_ds:
            assert 'units' not in hf_ds['time_bounds'].ncattrs()
            assert 'calendar' not in hf_ds['time_bounds'].ncattrs()


def test_no_time_case(no_time_case):
    input_head_dir, output_head_dir = no_time_case
    for file_name in listdir(input_head_dir):
        with Dataset(f"{input_head_dir}/{file_name}", 'r') as hf_ds:
            assert 'time' not in hf_ds.variables
            assert 'time_bounds' not in hf_ds.variables


def test_multistep_case(multistep_case):
    input_head_dir, output_head_dir = multistep_case
    assert len(listdir(input_head_dir)) == SIMPLE_NUM_TEST_HIST_FILES
    for file_name in listdir(input_head_dir):
        with Dataset(f"{input_head_dir}/{file_name}", 'r') as hf_ds:
            assert len(hf_ds['time'].dimensions) == 1


def test_with_auxiliary_case(with_auxiliary_case):
    input_head_dir, output_head_dir = with_auxiliary_case
    assert len(listdir(input_head_dir)) == SIMPLE_NUM_TEST_HIST_FILES

    for file_name in listdir(input_head_dir):
        with Dataset(f"{input_head_dir}/{file_name}", 'r') as hf_ds:
            for var_index in range(SIMPLE_NUM_VARS):
                assert hf_ds[f"VAR_AUX_{var_index}"].dimensions == ('time',)


def test_fragmented_case(spatial_fragment_case):
    input_head_dir, output_head_dir = spatial_fragment_case
    assert len(listdir(input_head_dir)) == FRAGMENTED_NUM_TIMESTEPS*FRAGMENTED_NUM_LAT_FILES*FRAGMENTED_NUM_LON_FILES
    
    dim_hashes = []

    for file_name in listdir(input_head_dir):
        with Dataset(f"{input_head_dir}/{file_name}", 'r') as hf_ds:
            for var_index in range(SIMPLE_NUM_VARS):
                assert hf_ds["lat"].size == FRAGMENTED_NUM_LAT_PTS_PER_HF
                assert hf_ds["lon"].size == FRAGMENTED_NUM_LON_PTS_PER_HF
                assert np.unique(hf_ds["lat"][:]).size == hf_ds["lat"].size
                assert np.unique(hf_ds["lon"][:]).size == hf_ds["lon"].size
            
            dim_hash = str([hf_ds["lat"][:], hf_ds["lon"][:]])
            if dim_hash not in dim_hashes:
                dim_hashes.append(dim_hash)
    
    assert len(dim_hashes) == FRAGMENTED_NUM_LAT_FILES*FRAGMENTED_NUM_LON_FILES


def test_mixed_timestep_case(mixed_timestep_case):
    input_head_dir, output_head_dir = mixed_timestep_case
    assert len(listdir(input_head_dir)) == MIXED_TS_NUM_TEST_HIST_FILES*4

    for index in range(MIXED_TS_NUM_TEST_HIST_FILES):
        with Dataset(f"{input_head_dir}/testing.hf0.{str(index).zfill(5)}.nc", 'r') as hf_ds:
            bounds = hf_ds["time_bounds"][:][0]
            assert 0 < bounds[1] - bounds[0] < 1
        with Dataset(f"{input_head_dir}/testing.hf1.{str(index).zfill(5)}.nc", 'r') as hf_ds:
            bounds = hf_ds["time_bounds"][:][0]
            assert 1 <= bounds[1] - bounds[0] < 28
        with Dataset(f"{input_head_dir}/testing.hf2.{str(index).zfill(5)}.nc", 'r') as hf_ds:
            bounds = hf_ds["time_bounds"][:][0]
            assert 28 <=bounds[1] - bounds[0] < 365
        with Dataset(f"{input_head_dir}/testing.hf3.{str(index).zfill(5)}.nc", 'r') as hf_ds:
            bounds = hf_ds["time_bounds"][:][0]
            assert 365 <= bounds[1] - bounds[0]


def test_auxiliary_only_case(auxiliary_only_case):
    input_head_dir, output_head_dir = auxiliary_only_case

    for file_name in listdir(input_head_dir):
        with Dataset(f"{input_head_dir}/{file_name}", 'r') as hf_ds:
            for var_index in range(SIMPLE_NUM_VARS):
                assert "time" == hf_ds[f"VAR_AUX_{var_index}"].dimensions[0]
                assert len(hf_ds[f"VAR_AUX_{var_index}"].dimensions) == 1
                assert f"VAR_{var_index}" not in hf_ds.variables


from gents.hfcollection import HFCollection
from gents.timeseries import TSCollection
from gents.mhfdataset import MHFDataset
from gents.tests.test_cases import generate_history_file
from gents.meta import get_attributes
import netCDF4
from os import listdir, makedirs

SIMPLE_SUITE_NUM_HIST_FILES = 100

class SimpleSuite:
    def setup(self):
        self.hf_head_dir = "hf/"
        self.ts_head_dir = "ts/"

        makedirs(self.hf_head_dir, exist_ok=True)
        makedirs(self.ts_head_dir, exist_ok=True)

        self.hf_paths = [f"{self.hf_head_dir}/benchmark.hf.{str(index).zfill(5)}.nc" for index in range(SIMPLE_SUITE_NUM_HIST_FILES)]

        for file_index, path in enumerate(self.hf_paths):
            generate_history_file(path, [(file_index+1)*30], [[file_index*30, (file_index+1)*30]])
    
    def time_hfcollection_create(self):
        hfc = HFCollection(self.hf_head_dir)

    def time_hfcollection_pull(self):
        hfc = HFCollection(self.hf_head_dir)
        hfc.pull_metadata()

    def time_tscollection_create(self):
        hfc = HFCollection(self.hf_head_dir)
        tsc = TSCollection(hfc, self.ts_head_dir)
    
    def time_tscollection_execute(self):
        hfc = HFCollection(self.hf_head_dir)
        tsc = TSCollection(hfc, self.ts_head_dir)
        tsc.execute()

    def time_mhfdataset(self):
        with MHFDataset(self.hf_paths) as agg_hf_ds:
            global_attrs = agg_hf_ds.get_global_attrs()
            data_vals = agg_hf_ds.get_var_vals("VAR0")
            assert data_vals is not None

    def time_nc_mfdataset(self):
        with netCDF4.MFDataset(self.hf_paths, aggdim="time") as agg_hf_ds:
            global_attrs = get_attributes(agg_hf_ds)
            data_vals = agg_hf_ds["VAR0"][:]
            assert data_vals is not None


#!/usr/bin/env python
"""
meta.py

Developer: Cameron Cummins
Contact: cameron.cummins@utexas.edu
Last Header Update: 07/03/25
"""
import netCDF4
import numpy as np
from cftime import num2date


def is_var_secondary(variable: netCDF4._netCDF4._Variable,
                     secondary_vars: list = ["time_bnds", "time_bnd", "time_bounds", "time_bound"],
                     secondary_dims: list = ["nbnd", "chars", "string_length", "hist_interval"],
                     max_num_dims: int = 1,
                     primary_dims: list = ["time"]) -> bool:
    """
    Determines if a variable is secondary or not (and then should be included in all time series files).
    Criteria are applied in order of the parameters described.

    :param dimensions: Dimensions for the variable.
    :param secondary_dims: List of secondary variable names
    :param secondary_dims: List of dimensions that make a variable secondary
    :param max_num_dims: Maximum number of dimensions a secondary variable should have.
    :param primary_dims: List of dimensions that make a variable primary (thus not secondary).
    :return: True if the variable is secondary, false if not.
    """
    if variable.name in secondary_vars:
        return True
        
    dims = np.unique(variable.dimensions)

    for tag in secondary_dims:
        if tag in dims:
            return True

    if len(dims) > max_num_dims:
        for tag in primary_dims:
            if tag in dims:
                return False

    return True


def get_attributes(dataset):
    """
    Builds Python dictionary of attributes from netCDF4 dataset and variable classes
    
    :param dataset: netCDF4 dataset or variable class object
    :return: Dictionary containing attributes.
    """
    attrs = {}
    if type(dataset) is netCDF4._netCDF4.MFDataset:
        for key in dataset.ncattrs():
            attrs[key] = dataset.__getattribute__(key)
    else:
        for key in dataset.ncattrs():
            attrs[key] = dataset.__getattr__(key)
    return attrs


def get_time_variables_names(ds):
    time_eqv = None
    time_bnds_eqv = None

    for var_name in ds.variables:
        lowercase_name = str(var_name).lower()
        if lowercase_name == 'time':
            time_eqv = var_name
        elif lowercase_name == 'time_bnds':
            time_bnds_eqv = var_name
        elif lowercase_name == 'time_bnd':
            time_bnds_eqv = var_name
        elif lowercase_name == 'time_bounds':
            time_bnds_eqv = var_name
        elif lowercase_name == 'time_bound':
            time_bnds_eqv = var_name
    
    return time_eqv, time_bnds_eqv


class netCDFMeta:
    """Stores and provides interface for accessing necessary metadata for individual history files."""
    def __init__(self, ds: netCDF4.Dataset):
        """
        :param ds: netCDF dataset read of history file (not this is not the path).
        """
        self.__time_vals = None
        self.__cftime_vals = None

        self.__attrs = get_attributes(ds)
        self.__path = ds.filepath()

        time_eqv, time_bnds_eqv = get_time_variables_names(ds)
        
        if time_eqv is None:
            raise ValueError(f"No equivalent time variable found to concatenate over. Path: {self.__path}")

        self.__time_vals = ds[time_eqv][:]

        if len(self.__time_vals.shape) > 1:
            self.__time_vals = np.squeeze(self.__time_vals)
        elif len(self.__time_vals.shape) == 0:
            self.__time_vals = np.array([self.__time_vals])

        if 'calendar' not in ds[time_eqv].ncattrs() or 'units' not in ds[time_eqv].ncattrs():
            raise AttributeError(f"Unable to pull 'calendar' and/or 'units' attributes from '{time_eqv}' time-equivalent variable. Path: {self.__path}")

        self.__cftime_vals = num2date(self.__time_vals, units=ds[time_eqv].units, calendar=ds[time_eqv].calendar)

        self.__time_bounds_vals = None
        self.__cftime_bounds_vals = None
        
        if time_bnds_eqv:
            self.__time_bounds_vals = ds[time_bnds_eqv][:]

            if len(self.__time_bounds_vals.shape) > 2:
                self.__time_bounds_vals = np.squeeze(self.__time_bounds_vals)
            elif len(self.__time_bounds_vals.shape) == 1:
                self.__time_bounds_vals = np.array([self.__time_bounds_vals])
            elif len(self.__time_bounds_vals.shape) == 0:
                raise ValueError(f"Found a 'time_bounds' equivalent variable, but it was a single value. It must have two values (one for each boundary). Path: {self.__path}")

            try:
                self.__cftime_bounds_vals = num2date(self.__time_bounds_vals, units=ds[time_bnds_eqv].units, calendar=ds[time_bnds_eqv].calendar)
            except AttributeError:
                self.__cftime_bounds_vals = num2date(self.__time_bounds_vals, units=ds[time_eqv].units, calendar=ds[time_eqv].calendar)
        self.__var_names = list(ds.variables)
        self.__primary_var_names = []
        self.__secondary_var_names = []
        
        for variable in ds.variables:
            if is_var_secondary(ds[variable]):
                self.__secondary_var_names.append(variable)
            else:
                self.__primary_var_names.append(variable)

        self.__dim_bounds = {}

        for dim_variable in ds.dimensions:
            if dim_variable in ds.variables:
                dim_data = ds[dim_variable][:]
                if dim_data.shape[0] >= 2:
                    self.__dim_bounds[dim_variable] = [np.min(dim_data), np.max(dim_data)]
                else:
                    self.__dim_bounds[dim_variable] = [np.min(dim_data)]

    def get_path(self):
        """
        :return: Get path to history file represented by this metadata class.
        """
        return self.__path
    
    def get_cftime_bounds(self):
        """
        :return: Get time bounds variable as CFTime objects.
        """
        return self.__cftime_bounds_vals

    def get_float_time_bounds(self):
        """
        :return: Get time bounds variable as floats.
        """
        return self.__time_bounds_vals

    def get_float_times(self):
        """
        :return: Get array of floats from time dimension.
        """
        return self.__time_vals

    def get_cftimes(self):
        """
        :return: Get array of CFTime objects from time dimension.
        """
        return self.__cftime_vals

    def get_variables(self):
        """
        :return: Get array of CFTime objects
        """
        return self.__var_names

    def get_primary_variables(self):
        """
        :return: Get primary variables derived from this history file.
        """
        return self.__primary_var_names

    def get_secondary_variables(self):
        """
        :return: Get secondary variables derived from this history file.
        """
        return self.__secondary_var_names

    def get_attributes(self):
        """
        :return: Get all attributes from this history file.
        """
        return self.__attrs

    def is_valid(self):
        """
        :return: Whether or not all necessary information is available for GenTS to create a time series.
        """
        if self.get_cftime_bounds() is None and self.get_cftimes() is None:
            return False
        elif len(self.get_primary_variables()) + len(self.get_secondary_variables()) == 0:
            return False
        elif "gents_version" in self.get_attributes():
            return False
        return True

    def get_dim_bounds(self):
        """
        :return: Dictionary containing the bounds for each dimension coordinate variable.
        """
        return self.__dim_bounds

def get_meta_from_path(path: str):
    """
    The netCDFMeta class only accepts netCDF4.Dataset objects as input to manage I/O errors.
    This function serves as a wrapper to obtain metadata.
    
    :param path: Path to netCDF file.
    :return: netCDF4 Dataset read of the history file.
    """
    ds_meta = None
    try:
        with netCDF4.Dataset(path, 'r') as ds:
            ds_meta = netCDFMeta(ds)
    except Exception as e:
        raise type(e)(f"{e} Path: {path}") from e

    return ds_meta


from gents.tests.test_cases import *
from gents.mhfdataset import *
from gents.hfcollection import HFCollection
from netCDF4 import Dataset
import numpy as np


def test_MHFDataset_simple(simple_case):
    input_head_dir, output_head_dir = simple_case
    hf_collection = HFCollection(input_head_dir)
    hf_groups = hf_collection.get_groups()

    for group in hf_groups:
        with MHFDataset(hf_groups[group]) as agg_hf_ds:
            assert len(agg_hf_ds) == len(hf_groups[group])
            for index in range(len(agg_hf_ds)):
                assert agg_hf_ds[index].filepath() == str(hf_groups[group][index])

            assert np.array_equal(agg_hf_ds.get_var_data_shape("VAR0"), (SIMPLE_NUM_TEST_HIST_FILES, 3, 4))

            var1_t0_output = agg_hf_ds.get_var_vals("VAR1", time_index_start=0, time_index_end=1)
            
            assert np.array_equal(var1_t0_output, np.ones(var1_t0_output.shape))


def test_MHFDataset_fragmented(spatial_fragment_case):
    input_head_dir, output_head_dir = spatial_fragment_case
    hf_collection = HFCollection(input_head_dir)
    hf_groups = hf_collection.get_groups()

    for group in hf_groups:
        with MHFDataset(hf_groups[group]) as agg_hf_ds:
            assert len(agg_hf_ds) == len(hf_groups[group])
            for index in range(len(agg_hf_ds)):
                assert agg_hf_ds[index].filepath() == str(hf_groups[group][index])

            assert np.array_equal(
                agg_hf_ds.get_var_data_shape("VAR0"),
                (
                    FRAGMENTED_NUM_TIMESTEPS,
                    FRAGMENTED_NUM_LAT_FILES*FRAGMENTED_NUM_LAT_PTS_PER_HF,
                    FRAGMENTED_NUM_LON_FILES*FRAGMENTED_NUM_LON_PTS_PER_HF
                )
            )


def test_get_concat_coords_simple(simple_case):
    input_head_dir, output_head_dir = simple_case
    hf_collection = HFCollection(input_head_dir)
    hf_groups = hf_collection.get_groups()

    for group in hf_groups:
        with MHFDataset(hf_groups[group]) as agg_hf_ds:
            coords = get_concat_coords(agg_hf_ds)
            for dim in coords:
                if dim == "time" or coords[dim] is None:
                    continue
                if dim in agg_hf_ds[0].variables:
                    assert np.array_equal(agg_hf_ds[0][dim][:], coords[dim])
                else:
                    assert agg_hf_ds[0].dimensions[dim].size == len(coords[dim])
            assert len(coords["time"]) == SIMPLE_NUM_TEST_HIST_FILES


def test_get_concat_coords_fragmented(spatial_fragment_case):
    input_head_dir, output_head_dir = spatial_fragment_case
    hf_collection = HFCollection(input_head_dir)
    hf_groups = hf_collection.get_groups()

    for group in hf_groups:
        with MHFDataset(hf_groups[group]) as agg_hf_ds:
            coords = get_concat_coords(agg_hf_ds)
            assert len(coords["lat"]) == FRAGMENTED_NUM_LAT_FILES*FRAGMENTED_NUM_LAT_PTS_PER_HF
            assert len(coords["lon"]) == FRAGMENTED_NUM_LON_FILES*FRAGMENTED_NUM_LON_PTS_PER_HF
            assert len(coords["time"]) == FRAGMENTED_NUM_TIMESTEPS


#!/usr/bin/env python
"""
timeseries.py

Developer: Cameron Cummins
Contact: cameron.cummins@utexas.edu
Last Header Update: 01/31/25
"""
import netCDF4
import numpy as np
import fnmatch
from os.path import isfile
from os import remove, makedirs
from pathlib import Path
from gents.meta import get_attributes
from gents.mhfdataset import MHFDataset
from gents.utils import get_version, LOG_LEVEL_IO_WARNING, ProgressBar
import traceback
import logging
import copy

logger = logging.getLogger(__name__)

try:
    import dask
    DASK_INSTALLED = True
except ImportError:
    DASK_INSTALLED = False
    logger.debug("Dask not installed. Proceeding in serial.")


def check_timeseries_integrity(ts_path: str):
    """
    Checks integrity of time series netCDF file by confirming `gents_version` attribute.

    :param ts_path: Path to time series file.
    :return: True if `gents_version` attribute is found. False if not (suggesting possible corruption).
    """
    try:
        with netCDF4.Dataset(ts_path, mode="r") as ts_ds:
            attrs = get_attributes(ts_ds)
        if "gents_version" in attrs:
            return True
    except OSError:
        logger.log(LOG_LEVEL_IO_WARNING, f"Corrupt timeseries output: '{ts_path}'")
    return False


def generate_time_series_error_wrapper(**args):
    try:
        return generate_time_series(**args)
    except Exception as e:
        print("=====================================================\n")
        print(f"START of GenTS Argument Dump for {type(e)}\n")
        print("=====================================================")
        for entry in args:
            print(f"\nArgument: '{entry}' \n")
            print(args[entry])
        print(f"\nError: '{type(e)}' \n")
        print(e)
        traceback.print_exc()
        print("=====================================================\n")
        print(f"END of GenTS Argument Dump for {type(e)}\n")
        print("=====================================================")
        raise type(e)(f"{e}") from e


def generate_time_series(hf_paths, ts_path_template, primary_var, secondary_vars, complevel=0, compression=None, overwrite=False, reference_structure=None):
    """
    Creates timeseries dataset from specified history file paths.

    :param hf_paths: List of paths to history files to generate time series from
    :param ts_out_dir: Directory to output time series files to.
    :param prefix: Prefix to add to beginning of the names of the generated time series files.
    :param complevel: Compression level to apply through netCDF4 API.
    :param compression: Compression algorithm to use through netCDF4 API.
    :param overwrite: Whether or not to delete existing time series files with the same names as those being generated.
    :param target_variable: Primary variable to extract from history files.
    :return: List of paths to time series generated.
    """

    ts_out_path = None
    with MHFDataset(hf_paths) as agg_hf_ds:
        global_attrs = agg_hf_ds.get_global_attrs()
        secondary_vars_data = {}
        
        for variable in secondary_vars:
            secondary_vars_data[variable] = agg_hf_ds.get_var_vals(variable)
        
        ts_string = agg_hf_ds.get_timestamp_string()

        ts_out_path = f"{ts_path_template}.{primary_var}.{ts_string}.nc"

        if overwrite and isfile(ts_out_path):
            remove(ts_out_path)
        elif not overwrite and isfile(ts_out_path):
            if check_timeseries_integrity(ts_out_path):
                return ts_out_path
            else:
                remove(ts_out_path)

        with netCDF4.Dataset(ts_out_path, mode="w") as ts_ds:
            if primary_var != "auxiliary":
                var_shape = agg_hf_ds.get_var_data_shape(primary_var)
                var_dims = agg_hf_ds.get_var_dimensions(primary_var)
                for index, dim in enumerate(var_dims):
                    if dim == "time":
                        ts_ds.createDimension(dim, None)
                    else:
                        ts_ds.createDimension(dim, var_shape[index])

                var_data = ts_ds.createVariable(primary_var,
                                                agg_hf_ds.get_var_dtype(primary_var),
                                                var_dims,
                                                complevel=complevel,
                                                compression=compression)
                var_data.set_auto_mask(False)
                var_data.set_auto_scale(False)
                var_data.set_always_mask(False)
                
                ts_ds[primary_var].setncatts(agg_hf_ds.get_var_attrs(primary_var))

                time_chunk_size = 1
                if len(var_shape) > 0 and "time" in var_dims:
                    for i in range(0, var_shape[0], time_chunk_size):
                        if i + time_chunk_size > var_shape[0]:
                            time_chunk_size = var_shape[0] - i
                        var_data[i:i + time_chunk_size] = agg_hf_ds.get_var_vals(
                            primary_var, time_index_start=i, time_index_end=i+time_chunk_size
                        )
                else:
                    var_data[:] = agg_hf_ds.get_var_vals(primary_var)

            for secondary_var in secondary_vars_data:
                var_shape = agg_hf_ds.get_var_data_shape(secondary_var)
                var_dims = agg_hf_ds.get_var_dimensions(secondary_var)

                for index, dim in enumerate(var_dims):
                    if dim not in ts_ds.dimensions:
                        if dim == "time":
                            ts_ds.createDimension(dim, None)
                        else:
                            ts_ds.createDimension(dim, var_shape[index])
                
                svar_data = ts_ds.createVariable(secondary_var,
                                                agg_hf_ds.get_var_dtype(secondary_var),
                                                var_dims,
                                                complevel=complevel,
                                                compression=compression)
                
                svar_data.set_auto_mask(False)
                svar_data.set_auto_scale(False)
                svar_data.set_always_mask(False)

                ts_ds[secondary_var].setncatts(agg_hf_ds.get_var_attrs(secondary_var))
                svar_data[:] = secondary_vars_data[secondary_var]
            
            ts_ds.setncatts(global_attrs | {"gents_version": str(get_version())})
    return ts_out_path


class TSCollection:
    """Time Series Collection that faciliates the creation of time series from a HFCollection."""
    def __init__(self, hf_collection, output_dir, ts_orders=None, dask_client=None):
        """
        :param hf_collection: History file collection to derive time series from
        :param output_dir: Directory to output time series files to
        :param ts_orders: List of Dask delayed functions of generate_time_series
        :param dask_client: Dask client to use when executing time series batches (Default: global client).
        """
        if dask_client is None and DASK_INSTALLED:
            self.__dask_client = dask.distributed.client._get_global_client()
        else:
            self.__dask_client = dask_client
        
        hf_collection.sort_along_time()

        self.__hf_collection = hf_collection
        self.__groups = self.__hf_collection.get_groups()
        self.__output_dir = output_dir
        
        if ts_orders is None:
            self.__hf_collection.pull_metadata()
            self.__orders = []
            for glob_template in self.__groups:
                output_template = glob_template.split(str(self.__hf_collection.get_input_dir()))[1]
                if "[sorting_pivot]" in output_template:
                    output_template = output_template.split("[sorting_pivot]")[0]
                ts_path_template = f"{self.__output_dir}{output_template}"
                hf_paths = self.__groups[glob_template]

                primary_vars = self.__hf_collection[hf_paths[0]].get_primary_variables()
                secondary_vars = self.__hf_collection[hf_paths[0]].get_secondary_variables()

                if len(primary_vars) > 0:
                    for var in primary_vars:
                        self.__orders.append({
                            "hf_paths": hf_paths,
                            "ts_path_template": ts_path_template[:-1],
                            "primary_var": var,
                            "secondary_vars": secondary_vars
                        })
                else:
                    self.__orders.append({
                        "hf_paths": hf_paths,
                        "ts_path_template": ts_path_template[:-1],
                        "primary_var": "auxiliary",
                        "secondary_vars": secondary_vars
                    })

            logger.debug(f"TSCollection initialized at '{output_dir}'.")
            logger.debug(f"{len(self.__orders)} timeseries orders generated.")
        else:
            self.__orders = ts_orders

    def __contains__(self, key):
        return key in self.__orders

    def __iter__(self):
        return iter(self.__orders)

    def __getitem__(self, index):
        return self.__orders[index]

    def __len__(self):
        return len(self.__orders)

    def items(self):
        return self.__orders.items()

    def values(self):
        return self.__orders.values()
    
    def get_hf_collection(self):
        return self.__hf_collection
    
    def copy(self, hf_collection=None, output_dir=None, ts_orders=None, dask_client=None):
        """
        Copies data of this TSCollection into a new one.
    
        :param hf_collection: HFCollection to assign to copy (defaults to existing).
        :param output_dir: Head output directory to assign to copy (defaults to existing).
        :param ts_orders: Time series orders to assign to copy (defaults to existing).
        :param dask_client: Dask client to assign to copy (defaults to existing).
        :return: TSCollection that is a copy.
        """
        if hf_collection is None:
            hf_collection = self.__hf_collection
        if output_dir is None:
            output_dir = self.__output_dir
        if ts_orders is None:
            ts_orders = self.__orders
        if dask_client is None:
            dask_client = self.__dask_client

        return TSCollection(hf_collection=hf_collection, output_dir=output_dir, ts_orders=ts_orders, dask_client=dask_client)

    def include(self, path_glob, var_glob="*"):
        """
        Applies inclusive filter to time series orders.

        :param path_glob: Glob pattern to apply to source history files.
        :param var_glob: Glob pattern to apply to primary variable names. Defaults to "*"
        :return: A new TSCollection that only includes time series orders that match the filter.
        """
        filtered_orders = []
        for order_dict in copy.deepcopy(self.__orders):
            path_matched = False
            for path in order_dict["hf_paths"]:
                if fnmatch.fnmatch(path, path_glob):
                    path_matched = True
                    break
            
            if path_matched and fnmatch.fnmatch(order_dict["primary_var"], var_glob):
                filtered_orders.append(order_dict)
        logger.debug(f"Inclusive filter(s) applied: '{var_glob}' to history files matching '{path_glob}'")
        return self.copy(ts_orders=filtered_orders)

    def exclude(self, path_glob, var_glob=""):
        """
        Applies exclusive filter to time series orders.

        :param path_glob: Glob pattern to apply to source history files.
        :param var_glob: Glob pattern to apply to primary variable names. Defaults to ""
        :return: A new TSCollection that excludes time series orders that match the filter.
        """
        filtered_orders = []
        for order_dict in copy.deepcopy(self.__orders):
            path_unmatched = True
            for path in order_dict["hf_paths"]:
                if fnmatch.fnmatch(path, path_glob):
                    path_unmatched = False
                    break
            
            if path_unmatched and not fnmatch.fnmatch(order_dict["primary_var"], var_glob):
                filtered_orders.append(order_dict)
        logger.debug(f"Exclusive filter(s) applied: '{var_glob}' to history files matching '{path_glob}'")
        return self.copy(ts_orders=filtered_orders)

    def add_args(self, path_glob="*", var_glob="*", level=None, alg=None, overwrite=None):
        """
        Applies arguments to pass to generate_time_series when processing time series orders.
        Filters specify which time series orders should be updated. If value is None, then the
        argument is not changed.

        :param path_glob: Glob pattern to match to source history files. Defaults to "*".
        :param var_glob: Glob pattern to match to primary variable names. Defaults to "*".
        :param level: Level of compresison to pass to the netCDF4 backend. Defaults to None.
        :param alg: Compression algorithm to pass to the netCDF4 backend. Defaults to None.
        :param overwrite: Whether or not to overwrite a time series output file if it already exists. Defaults to None.
        :return: A new TSCollection that includes time series orders with arguments applied.
        """
        new_orders = []
        for order_dict in copy.deepcopy(self.__orders):
            path_matched = False
            for path in order_dict["hf_paths"]:
                if fnmatch.fnmatch(path, path_glob):
                    path_matched = True
                    break
            
            if path_matched and fnmatch.fnmatch(order_dict["primary_var"], var_glob):
                if level is not None:
                    order_dict["complevel"] = level
                if alg is not None:
                    order_dict["compression"] = alg
                if overwrite is not None:
                    order_dict["overwrite"] = overwrite
            new_orders.append(order_dict)

        logger.debug(f"Arguments applied (excluding None): ['level': {level}, 'alg': {alg}, 'overwrite': {overwrite}] to history files matching '{path_glob}' and variables matching '{var_glob}'.")
        return self.copy(ts_orders=new_orders)

    def apply_path_swap(self, string_match, string_swap, path_glob="*", var_glob="*"):
        """
        Iterates over time series output path templates, finds the ones that mach the filter, and
        replaces the matching string with a swap string if it exists.
    
        :param path_glob: Glob pattern to apply to source history files. Defaults to "*"
        :param var_glob: Glob pattern to apply to primary variable names. Defaults to "*".
        :param string_match: String to match to in output template path.
        :param string_swap: String to replace match with, if found.
        :return: A new TSCollection with updated output path templates
        """
        new_orders = []
        for order_dict in copy.deepcopy(self.__orders):
            for path in order_dict["hf_paths"]:
                if fnmatch.fnmatch(path, path_glob):
                    order_dict["ts_path_template"] = order_dict["ts_path_template"].replace(string_match, string_swap)
            new_orders.append(order_dict)
    
        logger.debug(f"Path swap '{string_match}' -> '{string_swap}' to history files matching '{path_glob}' and variables matching '{var_glob}'.")
        return self.copy(ts_orders=new_orders)
        
    def apply_compression(self, level, alg, path_glob, var_glob="*"):
        """
        Applies compression arguments to time series orders.

        :param level: Level of compresison to pass to the netCDF4 backend.
        :param alg: Compression algorithm to pass to the netCDF4 backend.
        :param path_glob: Glob pattern to match to source history files.
        :param var_glob: Glob pattern to match to primary variable names. Defaults to "*".
        :return: A new TSCollection that includes time series orders with arguments applied.
        """
        return self.add_args(path_glob=path_glob, var_glob=var_glob, level=level, alg=alg)

    def apply_overwrite(self, path_glob, var_glob="*"):
        """
        Applies overwrite argument to time series orders.

        :param path_glob: Glob pattern to match to source history files.
        :param var_glob: Glob pattern to match to primary variable names. Defaults to "*".
        :return: A new TSCollection that includes time series orders with arguments applied.
        """
        return self.add_args(path_glob=path_glob, var_glob=var_glob, overwrite=True)

    def append_timestep_dirs(self, var_glob="*"):
        """
        Appends directories named according to time-step to the end of the timeseries output path templates.
        This sorts the output into bins by the time-step frequency (i.e. hour_1, day_1, month_1, year_1)

        :param var_glob: Glob pattern to match to primary variable names. Defaults to "*".
        :return: A new TSCollection with time-step directories added.
        """
        new_orders = []
        for order_dict in copy.deepcopy(self.__orders):
            if fnmatch.fnmatch(order_dict["primary_var"], var_glob):
                paths = order_dict["hf_paths"]
                paths.sort()
                dt = None
                times = self.__hf_collection[paths[0]].get_cftimes()
                if len(times) > 1:
                    dt = (times[1] - times[1])
                elif len(paths) > 1:
                    time_0 = self.__hf_collection[paths[0]].get_cftimes()[0]
                    time_1 = self.__hf_collection[paths[1]].get_cftimes()[0]
                    dt = time_1 - time_0

                if dt is None:
                    timestep_label = "unsorted"
                elif dt.days == 0:
                    timestep_label = "hour_1"
                elif dt.days < 28:
                    timestep_label = "day_1"
                elif dt.days < 365:
                    timestep_label = "month_1"
                else:
                    timestep_label = "year_1"

                template = Path(order_dict["ts_path_template"])
                order_dict["ts_path_template"] = str(template.parent) + f"/{timestep_label}/" + template.name

                new_orders.append(order_dict)
        return self.copy(ts_orders=new_orders)

    def remove_overwrite(self, path_glob, var_glob="*"):
        """
        Removes overwrite argument to time series orders.

        :param path_glob: Glob pattern to match to source history files.
        :param var_glob: Glob pattern to match to primary variable names. Defaults to "*".
        :return: A new TSCollection that includes time series orders with arguments applied.
        """
        return self.add_args(path_glob=path_glob, var_glob=var_glob, overwrite=False)

    def get_dask_delayed(self):
        """Gets list of delayed time series generation functions."""
        if DASK_INSTALLED:
            delayed_orders = []
            for args in self.__orders:
                delayed_orders.append(dask.delayed(generate_time_series_error_wrapper)(**args))
            return delayed_orders
        else:
            raise ImportError("Dask not installed!")

    def create_directories(self, exist_ok=True):
        """Creates directory structure to output time series files to."""
        logger.info("Creating directory structure for time series output.")
        for order_dict in self.__orders:
            makedirs(Path(order_dict['ts_path_template']).parent, exist_ok=exist_ok)

    def execute(self):
        """Execute delayed time series generation functions across the Dask cluster."""
        self.create_directories()
        results = []
        if self.__dask_client is None:
            logger.info("No Dask client detected... proceeding in serial.")
            prog_bar = ProgressBar(total=len(self.__orders))
            for args in self.__orders:
                results.append(generate_time_series_error_wrapper(**args))
                prog_bar.step()
        else:
            logger.info("Dask client detected! Generating time series files in parallel.")
            results = self.__dask_client.compute(self.get_dask_delayed(), sync=True)
        return results


from gents.tests.test_cases import *
from gents.hfcollection import *
from gents.meta import netCDFMeta
from gents.utils import enable_logging
from pathlib import PosixPath
from os.path import isfile
import numpy as np


def test_find_files(structured_case):
    input_head_dir, output_head_dir = structured_case
    num_files = STRUCTURED_NUM_VARS*STRUCTURED_NUM_DIRS*STRUCTURED_NUM_SUBDIRS
    assert len(find_files(input_head_dir, "*.nc")) == num_files

    with open(f'{input_head_dir}/empty_file.txt', 'w') as f: pass
    assert len(find_files(input_head_dir, "*.nc")) == num_files
    assert len(find_files(input_head_dir, "*.txt")) == 1
    assert len(find_files(input_head_dir, "*")) == 1 + num_files


def test_calculate_year_slices():
    assert calculate_year_slices(10, 0, 30) == [(0, 9), (10, 19), (20, 29), (30, 39)]
    assert calculate_year_slices(1, 0, 3) == [(0, 0), (1, 1), (2, 2), (3, 3)]
    assert calculate_year_slices(5, 3, 12) == [(3, 7), (8, 12)]


def test_hf_sorting(structured_case):
    input_head_dir, output_head_dir = structured_case
    hf_paths = find_files(input_head_dir, "*.nc")
    groups = sort_hf_groups(hf_paths)

    num_files = 0
    for group in groups:
        num_files += len(groups[group])

    assert len(groups) == STRUCTURED_NUM_DIRS*STRUCTURED_NUM_SUBDIRS
    assert num_files == len(hf_paths)

    new_paths = hf_paths + [PosixPath(f"{path.parent}/{path.name.replace('testing', 'other_testing')}") for path in hf_paths]
    groups = sort_hf_groups(new_paths)

    num_files = 0
    for group in groups:
        num_files += len(groups[group])

    assert len(groups) == 2*STRUCTURED_NUM_DIRS*STRUCTURED_NUM_SUBDIRS
    assert num_files == 2*len(hf_paths)


def test_get_year_bounds(simple_case, scrambled_case, structured_case):
    input_head_dir, output_head_dir = simple_case
    hf_collection = HFCollection(input_head_dir)
    hf_collection.pull_metadata()
    assert get_year_bounds(hf_collection) == (CASE_START_YEAR, int(CASE_START_YEAR+((SIMPLE_NUM_TEST_HIST_FILES-1)/12)))

    input_head_dir, output_head_dir = scrambled_case
    hf_collection = HFCollection(input_head_dir)
    hf_collection.pull_metadata()
    
    assert get_year_bounds(hf_collection) == (CASE_START_YEAR, int(CASE_START_YEAR+((SCRAMBLED_NUM_TEST_HIST_FILES-1)/12)))

    input_head_dir, output_head_dir = structured_case
    hf_collection = HFCollection(input_head_dir)
    hf_collection.pull_metadata()
    assert get_year_bounds(hf_collection) == (CASE_START_YEAR, int(CASE_START_YEAR+((STRUCTURED_NUM_TEST_HIST_FILES-1)/12)))


def test_simple_hfcollection(simple_case, caplog):
    input_head_dir, output_head_dir = simple_case
    hf_collection = HFCollection(input_head_dir)
    
    assert hf_collection.get_input_dir() == input_head_dir
    assert hasattr(hf_collection, '__iter__')
    assert len(hf_collection) == SIMPLE_NUM_TEST_HIST_FILES
    for path in hf_collection:
        assert isfile(path)


    excluded = hf_collection.exclude(["*.00001.nc"])
    included = hf_collection.include(["*.00001.nc"])
    
    assert len(excluded) == SIMPLE_NUM_TEST_HIST_FILES - 1
    assert len(included) == 1
    assert hf_collection is not excluded
    assert hf_collection is not included

    assert not hf_collection.is_pulled()

    hf_collection.pull_metadata()

    assert hf_collection.is_pulled()

    meta_data_pulled = False
    for path in hf_collection:
        if hf_collection[path] is not None:
            meta_data_pulled = True
            break

    valid_checks = hf_collection.check_validity()
    assert type(valid_checks) is dict
    assert len(valid_checks) == 0
    
    path_sample = list(hf_collection)[0]
    meta_sample = hf_collection[path_sample]
    assert meta_sample.get_path() == str(path_sample)
    assert type(meta_sample) == netCDFMeta
    
    groups = hf_collection.get_groups()
    assert type(groups) == dict
    assert len(groups) == 1
    assert len(groups[list(groups)[0]]) == len(hf_collection)

    group_meta_map = {path: hf_collection[path] for path in hf_collection}
    
    min_year, max_year = get_year_bounds(group_meta_map)
    assert (min_year, max_year) == (CASE_START_YEAR, CASE_START_YEAR + int(np.floor(SIMPLE_NUM_TEST_HIST_FILES / 12)))
    
    for slice_size in range(1, 10):
        repeat_years = []
        year_slices = calculate_year_slices(slice_size, min_year, max_year)
        for lower, upper in year_slices:
            assert upper - lower <= slice_size
            assert lower <= upper
            assert lower not in repeat_years
            assert upper not in repeat_years
            repeat_years.append(lower)
            repeat_years.append(upper)
        assert year_slices[0][0] <= min_year
        assert year_slices[-1][1] >= max_year

    sliced_groups = hf_collection.slice_groups(slice_size_years=1).get_groups()
    assert len(sliced_groups) == int(np.ceil(SIMPLE_NUM_TEST_HIST_FILES / 12))
    
    for group in list(sliced_groups)[:-1]:
        assert len(sliced_groups[group]) == 12

    assert len(caplog.text) == 0


def test_time_bounds_case(time_bounds_case):
    input_head_dir, output_head_dir = time_bounds_case
    hf_collection = HFCollection(input_head_dir)
    hf_collection.pull_metadata()

    assert len(hf_collection) == TIME_NUM_TEST_HIST_FILES


def test_no_times_case(no_time_case):
    input_head_dir, output_head_dir = no_time_case
    hf_collection = HFCollection(input_head_dir)
    with pytest.raises(ValueError, match=".nc"):
        hf_collection.pull_metadata()


def test_scrambled_hfcollection(scrambled_case):
    input_head_dir, output_head_dir = scrambled_case
    hf_collection = HFCollection(input_head_dir)

    assert len(hf_collection) == SCRAMBLED_NUM_TEST_HIST_FILES

    hf_collection.pull_metadata()
    assert len(hf_collection.check_validity()) == 0


def test_structured_hfcollection(structured_case):
    input_head_dir, output_head_dir = structured_case
    hf_collection = HFCollection(input_head_dir)

    assert len(hf_collection) == STRUCTURED_NUM_TEST_HIST_FILES*STRUCTURED_NUM_DIRS*STRUCTURED_NUM_SUBDIRS

    hf_collection.pull_metadata()
    assert len(hf_collection.check_validity()) == 0


def test_hfcollection_copy(simple_case):
    input_head_dir, output_head_dir = simple_case
    hf_collection = HFCollection(input_head_dir)

    hf_copy = hf_collection.copy()
    assert type(hf_copy) == HFCollection
    assert list(hf_copy) == list(hf_collection)
    assert hf_copy is not hf_collection

    hf_copy = hf_collection.include("*.nc")
    assert type(hf_copy) == HFCollection
    assert list(hf_copy) == list(hf_collection)
    assert hf_copy is not hf_collection

    hf_copy = hf_collection.exclude("*.txt")
    assert type(hf_copy) == HFCollection
    assert list(hf_copy) == list(hf_collection)
    assert hf_copy is not hf_collection
    
    hf_copy = hf_collection.include_years(0, 99999)
    assert type(hf_copy) == HFCollection
    assert list(hf_copy) == list(hf_collection)
    assert hf_copy is not hf_collection
    
    hf_copy = hf_collection.slice_groups(1)
    assert type(hf_copy) == HFCollection
    assert list(hf_copy) == list(hf_collection)
    assert hf_copy is not hf_collection


def test_missing_time_bounds_attrs(simple_case_missing_attrs):
    input_head_dir, output_head_dir = simple_case_missing_attrs
    hf_collection = HFCollection(input_head_dir)
    hf_collection.pull_metadata()

    for entry in hf_collection:
        assert hf_collection[entry].is_valid()


def test_spatially_fragmented_hf(spatial_fragment_case):
    input_head_dir, output_head_dir = spatial_fragment_case
    hf_collection = HFCollection(input_head_dir)
    hf_collection.pull_metadata()

    assert len(hf_collection) == FRAGMENTED_NUM_TIMESTEPS*FRAGMENTED_NUM_LAT_FILES*FRAGMENTED_NUM_LON_FILES

    groups = hf_collection.get_groups(check_fragmented=True)

    assert len(groups) == 1


from gents.hfcollection import HFCollection
from gents.timeseries import TSCollection
from gents.utils import get_version
from gents.tests.test_cases import *
from gents.tests.test_workflow import is_monotonic
from netCDF4 import Dataset
from pytest import importorskip


def test_dask_simple_workflow(simple_case):
    daskd = importorskip("dask.distributed", reason="Dask distributed not installed.")
    input_head_dir, output_head_dir = simple_case

    cluster = daskd.LocalCluster(n_workers=2, threads_per_worker=1, memory_limit="2GB")
    client = cluster.get_client()

    hf_collection = HFCollection(input_head_dir, dask_client=client)
    ts_collection = TSCollection(hf_collection, output_head_dir, dask_client=client)
    ts_paths = ts_collection.execute()

    hf_collection.sort_along_time()

    assert len(ts_paths) == SIMPLE_NUM_VARS
    
    for path in ts_paths:
        with Dataset(path, 'r') as ts_ds:
            assert ts_ds["time"].size == SIMPLE_NUM_TEST_HIST_FILES
            assert ts_ds.getncattr("gents_version") == get_version()
            var_name = path.split(".")[-3]
            
            for index in range(ts_ds["time"].size):
                with Dataset(list(hf_collection)[index], 'r') as hf_ds:
                    assert (ts_ds[var_name][index] == hf_ds[var_name]).all()

                    for key in hf_ds.ncattrs():
                        assert ts_ds.getncattr(key) == hf_ds.getncattr(key)
                    
                    for key in hf_ds[var_name].ncattrs():
                        assert ts_ds[var_name].getncattr(key) == hf_ds[var_name].getncattr(key)
    client.shutdown()


def test_dask_scrambled_workflow(scrambled_case):
    daskd = importorskip("dask.distributed", reason="Dask distributed not installed.")
    input_head_dir, output_head_dir = scrambled_case

    cluster = daskd.LocalCluster(n_workers=2, threads_per_worker=1, memory_limit="2GB")
    client = cluster.get_client()

    hf_collection = HFCollection(input_head_dir, dask_client=client)
    ts_collection = TSCollection(hf_collection, output_head_dir, dask_client=client)
    ts_paths = ts_collection.execute()

    hf_collection.sort_along_time()

    assert len(ts_paths) == SCRAMBLED_NUM_VARS

    for path in ts_paths:
        with Dataset(path, 'r') as ts_ds:
            assert ts_ds["time"].size == SCRAMBLED_NUM_TEST_HIST_FILES
            assert is_monotonic(ts_ds["time"][:])

    client.shutdown()


def test_dask_structured_workflow(structured_case):
    daskd = importorskip("dask.distributed", reason="Dask distributed not installed.")
    input_head_dir, output_head_dir = structured_case

    cluster = daskd.LocalCluster(n_workers=2, threads_per_worker=1, memory_limit="2GB")
    client = cluster.get_client()

    hf_collection = HFCollection(input_head_dir, dask_client=client)
    ts_collection = TSCollection(hf_collection, output_head_dir, dask_client=client)
    ts_paths = ts_collection.execute()
    
    assert len(ts_paths) == SCRAMBLED_NUM_VARS*STRUCTURED_NUM_DIRS*STRUCTURED_NUM_SUBDIRS

    client.shutdown()



#!/usr/bin/env python
"""
hfcollection.py

Developer: Cameron Cummins
Contact: cameron.cummins@utexas.edu
Last Header Update: 04/30/25
"""
from gents.meta import get_meta_from_path
from gents.utils import ProgressBar, LOG_LEVEL_IO_WARNING
from cftime import num2date
from pathlib import Path
import numpy as np
import os
import fnmatch
import cftime
import netCDF4
import warnings
import logging

logging.captureWarnings(True)
logger = logging.getLogger(__name__)

try:
    from dask.distributed import client
    import dask
    DASK_INSTALLED = True
except ImportError:
    DASK_INSTALLED = False
    logger.debug("Dask not installed. Proceeding in serial.")


def check_config(config):
    """
    Used to ensure configuration dictionary has necessary parameters.

    :param config: The config directory to check.
    :return: True if config dictionary is correctly set, false if not.
    """
    assert "name" in config
    assert "include" in config
    assert "exclude" in config
    assert type(config["include"]) is dict or config["include"] is None
    assert type(config["exclude"]) is list or config["exclude"] is None


def get_default_config():
    """
    Generates dictionary containing default configurations for GenTS.

    :return: Dictionary with default parameters.
    """
    return {
        "name": "default",
        "include": None,
        "exclude": None
    }


def find_files(head_path, pattern):
    """
    Search for files in the specified head directory and all subdirectories that match the given wildcard pattern.

    :param head_path: The head directory to start searching from.
    :param pattern: The wildcard pattern to match files against (e.g., '*.nc').
    :return: A list of file paths that match the pattern.
    """
    matched_files = []

    for root, dirs, files in os.walk(head_path):
        for file in files:
            if fnmatch.fnmatch(file, pattern):
                matched_files.append(Path(os.path.join(root, file)))

    return sorted(matched_files)


def calculate_year_slices(slice_size_years, min_year, max_year):
    """
    Calculates the ranges for each slice in years within a given range.

    :param slice_size_years: Length of each slice in years.
    :param min_year: Minimum or starting year for the full range.
    :param max_year: Maximum or ending year for the full range.
    :return: List of tuples where each tuple defines the year range for each slice
    """
    if max_year < min_year:
        raise ValueError("Maximum year cannot exceed minimum year.")
    if slice_size_years >= max_year - min_year:
        return [(min_year, max_year)]

    start_year = min_year
    end_year = int(np.ceil(max_year / slice_size_years)*slice_size_years)

    ranges = []
    for year in np.arange(start_year, end_year+slice_size_years, slice_size_years, dtype=int):
        ranges.append((int(year), int(year+slice_size_years-1)))
        if year >= max_year or year+slice_size_years-1 >= max_year:
            break
    
    return ranges


def find_all_indices(string, substring):
    """
    Finds all indices where the substring occurs in a string.

    :param string: The string to search in.
    :param substring: The substring to search for.
    :return: A list of indices where the substring is found.
    """
    indices = []
    start = 0
    while True:
        index = string.find(substring, start)
        if index == -1:
            break
        indices.append(index)
        start = index + 1
    return indices


def sort_hf_groups(hf_paths, delimiter=".", substring_index=2):
    """
    Groups history file paths by directory and substrings.

    :param hf_paths: List of paths to history files.
    :param delimiter: The delimiter to find the substrings with.
    :param substring_index: Substrings to skip after splitting with the delimiter (from right to left).
    :return: A dictionary that maps group substrings (key) to history file paths (value)
    """
    directory_groups = {}
    for path in hf_paths:
        if path.parent in directory_groups:
            directory_groups[path.parent].append(path)
        else:
            directory_groups[path.parent]= [path]

    hf_groups = {}
    for parent_path in directory_groups:
        group_paths = [path for path in directory_groups[parent_path]]
        substrings = []
        for path in group_paths:
            parsed = path.name[:find_all_indices(path.name, delimiter)[-1 * substring_index]]
            substrings.append(parsed)
        
        for substring in np.unique(substrings):
            hf_groups[f"{parent_path}/{substring}*"] = []
            for path in group_paths:
                parsed = path.name[:find_all_indices(path.name, delimiter)[-1 * substring_index]]
                if substring == parsed:
                    hf_groups[f"{parent_path}/{substring}*"].append(path)
        
    return hf_groups


def get_year_bounds(hf_to_meta_map):
    """
    Determines the minimum and maximum year for a series of mapped history files.

    :param hf_to_time_map: Dictionary that maps history file paths (key) to their time and time bound data (value)
    :return: The minimum and maximum years from the specified history files as a tuple
    """
    min_year = np.inf
    max_year = -np.inf
    
    for path in list(hf_to_meta_map.keys()):
        time_bounds = hf_to_meta_map[path].get_cftime_bounds()
        if time_bounds is None:
            time_bounds = []
            for ts in hf_to_meta_map[path].get_cftimes():
                time_bounds.append([ts, ts])
        
        for index in range(len(time_bounds)):
            lower_bound, upper_bound = time_bounds[index]
            
            mid_time = lower_bound + ((upper_bound - lower_bound) / 2)

            if mid_time.year > max_year:
                max_year = mid_time.year
            if mid_time.year < min_year:
                min_year = mid_time.year
    return min_year, max_year


def generate_output_template(hf_head_dir, group_path_id, output_head_dir=None, directory_swaps={"hist": "tseries"}, filename_delimiter=".", cutoff_index=None):
    """
    Creates file path template (missing file name suffix) for outputing timeseries files to.

    :param hf_head_dir: Head directory used to read in the history files.
    :param group_path_id: One group path ID produced by slice_hf_groups.
    :param output_head_dir: Head directory to output to, if not the same as the input head directory.
    :param directory_swaps: Dictionary with directory names (key) to rename (value)
    :param filename_delimiter: Delimiter used to separate file name tags
    :param cutoff_index: Index to cutoff tags that are generated using the delimiter (typically used to remove the date)
    :return: Output template path for generating time series files.
    """
    group_path_id = Path(group_path_id)
                         
    raw_filename_prefix = group_path_id.name
    if cutoff_index is None:
        cutoff_index = find_all_indices(raw_filename_prefix, ".")[-1]
    filename_prefix = raw_filename_prefix[:cutoff_index]
    
    sub_dir_structure = (str(group_path_id.parent).split(hf_head_dir)[-1]).split("/")

    for key in directory_swaps:
        for index in range(len(sub_dir_structure)):
            if sub_dir_structure[index] == key:
                sub_dir_structure[index] = directory_swaps[key]

    sub_dir_path = "/"
    for directory in sub_dir_structure:
        sub_dir_path += f"{directory}/"

    if output_head_dir is None:
        output_template = Path(f"{hf_head_dir}/{sub_dir_path}/{filename_prefix}")
    else:
        output_template = Path(f"{output_head_dir}/{sub_dir_path}/{filename_prefix}")
    return output_template


def is_ds_within_years(ds_meta, min_year, max_year):
    """
    Checks if time bounds from dataset metadata are between the year range.
    
    :param ds_meta: Metadata associated with the dataset.
    :param min_year: Minimum year in range.
    :param max_year: Maximum year in range.
    :return: True if time bounds are within the year range, false if not.
    """
    time_bounds = ds_meta.get_cftime_bounds()
    if time_bounds is not None:
        year = (time_bounds[0].year + time_bounds[1].year) / 2
    else:
        year = ds_meta.get_cftimes()[0]

    if min_year <= year <= max_year:
        return True
    else:
        return False


def filter_by_variables(meta_datasets):
    """
    Checks for consistency within a list of datasets.
    
    :param ds_meta: List of metadatas associated with the datasets.
    :return: Tuple containing majority variables
    """
    variable_sets = {}
    for index in range(len(meta_datasets)):
        var_set = meta_datasets[index].get_variables()
        var_set.sort()

        if str(var_set) in variable_sets:
            variable_sets[str(var_set)].append(index)
        else:
            variable_sets[str(var_set)] = [index]

    majority = None
    others = None
    
    if len(variable_sets) > 1:
        counts = []
        for var_set in variable_sets:
            counts.append(len(variable_sets[var_set]))

        majority_index = np.argmax(counts)
        if np.sum(counts[majority_index] == np.array(counts)) == 1:
            majority_set = list(variable_sets)[majority_index]
    
            majority = []
            others = []
            for index in range(len(meta_datasets)):
                if index in variable_sets[majority_set]:
                    majority.append(meta_datasets[index])
                else:
                    others.append(meta_datasets[index])
    else:
        majority = meta_datasets
    
    return (majority, others)


def sort_metas_by_time(metas):
    """
    Re-arranges the metadata objects in a list so that they are ordered by time.
    
    :param metas: List of unsorted metadata.
    :return: List of metadata sorted by time.
    """
    time_sorted_metas = [metas[0]]

    if len(metas) > 1:
        for meta in metas[1:]:
            meta_sorted = False
            
            for index in range(len(time_sorted_metas)):
                if meta.get_cftimes()[0] < time_sorted_metas[index].get_cftimes()[0]:
                    time_sorted_metas.insert(index, meta)
                    meta_sorted = True
                    break
            if not meta_sorted:
                time_sorted_metas.append(meta)

    return time_sorted_metas

    
def check_groups_by_variables(sliced_groups):
    """
    Checks if the history files for each group have the same sets of variables.
    If an inconsistent set is discovered, it is thrown out and a warning is logged.
    
    :param sliced_groups: Sliced groups of history files to check.
    :return: Sliced groups with inconsistent sets filtered out.
    """
    filtered_sliced_groups = {}
    for group in sliced_groups:
        meta_datasets = sliced_groups[group]
        majority, others = filter_by_variables(meta_datasets)
        if majority is not None:
            filtered_sliced_groups[group] = sort_metas_by_time(majority)
            if others is not None:
                for meta_ds in others:
                    logger.log(LOG_LEVEL_IO_WARNING, f"Dataset has inconsistent variable list with directory group: {meta_ds.get_path()}")
        else:
            logger.log(LOG_LEVEL_IO_WARNING, f"Unable to determine majority dataset, check variable configurations between directory groups, group ID: {group}")
    return filtered_sliced_groups


def merge_fragmented_groups(hf_groups, hf_meta_map):
    """
    Iterates through grouped history files and merges groups that have identical time values, variable names,
    but differing values on other dimensions.
    
    :param groups: Grouped history file dictionary to iterate through.
    :return: New grouped history file dictionary with fragmented groups merged together. 
    """
    new_groups = {}
    fragmented_groups = {}

    for pattern in hf_groups:
        init_path = str(hf_groups[pattern][0])
        if init_path[-3:] == '.nc':
            new_groups[pattern] = hf_groups[pattern]
        else:
            fragmented_groups[pattern] = hf_groups[pattern]

    if len(fragmented_groups) > 0:
        num_fragmented_files = sum([len(fragmented_groups[pattern]) for pattern in fragmented_groups])
        logger.info(f"Found {num_fragmented_files} spatially fragmented files.")

    dim_hashes = {}
    for pattern in fragmented_groups:
        dims = hf_meta_map[fragmented_groups[pattern][0]].get_dim_bounds()
        dims = {variable: dims[variable] for variable in dims if "time" not in dims}
        dims_hash = str(dims)

        if dims_hash not in dim_hashes:
            dim_hashes[dims_hash] = []
        
        for path in fragmented_groups[pattern]:
            dim_hashes[dims_hash].append(path)

    for dim_hash in dim_hashes:
        paths = dim_hashes[dim_hash]
        label = str(paths[0]).split(".nc")[0] + "*"
        if label not in new_groups:
            new_groups[label] = paths
        else:
            raise KeyError(f"Fragmentation merge failed! History file group '{label}' already exists. Try filtering to fragmented files only to minimize confusion with non-fragmented history files.")

    return new_groups


class HFCollection:
    """History File Collection, holds paths to all history files and serves as an interface for interpreting the metadata."""
    def __init__(self, hf_dir, dask_client=None, meta_map=None, hf_groups=None, hf_glob_pattern="*.nc*"):
        """
        :param hf_dir: Head directory to history files
        :param dask_client: Dask client object. If not given, the global client is used instead.
        :param meta_map: History file to metadata map to use if deriving from existing HFCollection (overrides recursive search with hf_dir).
        :param hf_groups: History file groups if deriving from existing HFCollection (overrides recursive search with hf_dir).
        :param hf_glob_pattern: Glob pattern to match files against when searching recursively through the head directory.
        """
        self.__raw_paths = find_files(hf_dir, hf_glob_pattern)

        if dask_client is None and DASK_INSTALLED:
            self.__client = dask.distributed.client._get_global_client()
        else:
            self.__client = dask_client

        self.__hf_to_meta_map = {}
        if meta_map is None:
            for path in self.__raw_paths:
                self.__hf_to_meta_map[path] = None
        else:
            self.__hf_to_meta_map = meta_map

        self.__meta_pulled = self.is_pulled()
        
        self.__hf_groups = hf_groups
        self.__hf_dir = hf_dir

        if meta_map is None and hf_groups is None:
            logger.info(f"Initialized HFCollection at '{hf_dir}'")
            logger.info(f"{len(self.__raw_paths)} netCDF files found.")

    def __getitem__(self, key):
        return self.__hf_to_meta_map[key]

    def __contains__(self, key):
        return key in self.__hf_to_meta_map

    def __iter__(self):
        return iter(self.__hf_to_meta_map)

    def __len__(self):
        return len(self.__hf_to_meta_map)

    def items(self):
        return self.__hf_to_meta_map.items()

    def values(self):
        return self.__hf_to_meta_map.values()

    def keys(self):
        return self.__hf_to_meta_map.keys()

    def is_pulled(self):
        """Returns true if all metadata has been pulled. False if not."""
        for path in self.__hf_to_meta_map:
            if self.__hf_to_meta_map[path] is None:
                return False
        return True

    def get_input_dir(self):
        """Return the input directory"""
        return self.__hf_dir

    def check_pulled(self):
        """Checks if metadata has been pulled. If not, then pull."""
        if not self.__meta_pulled:
            self.pull_metadata()

    def copy(self, dask_client=None, meta_map=None, hf_groups=None):
        """
        Copies data of this HFCollection into a new one.
    
        :param dask_client: Dask client to assign to copy.
        :param meta_map: history file to metadata map to use when copying (defaults to existing).
        :return: HFCollection that is a copy.
        """
        if dask_client is None:
            dask_client = self.__client
        if meta_map is None:
            meta_map = self.__hf_to_meta_map
        if hf_groups is None and self.__meta_pulled:
            hf_groups = self.get_groups()
        return HFCollection(self.__hf_dir, dask_client=dask_client, meta_map=meta_map, hf_groups=hf_groups)

    def sort_along_time(self):
        """
        Sorts the history files along the time dimension formed out of the time values pulled from metadata.
        """
        self.check_pulled()

        sorted_map = dict(sorted(self.__hf_to_meta_map.items(), key=lambda item: item[1].get_float_times()[0]))
        self.__hf_to_meta_map = sorted_map
        logger.info(f"Sorted along time.")
        return self.copy(meta_map=sorted_map)
    
    def pull_metadata(self, check_valid=True):
        """Pulls metadata associated with each history file in the collection."""
        logger.info(f"Pulling metadata...")
        ds_metas_futures = []
        ds_metas = []
        paths = list(self.__hf_to_meta_map.keys())

        if self.__client is None:
            prog_bar = ProgressBar(total=len(paths))
            for path in paths:
                ds_metas.append(get_meta_from_path(path))
                prog_bar.step()
        else:
            prog_bar = ProgressBar(total=np.min(1, len(paths) // 10000))
            for index in range(0, len(paths), 10000):
                ds_metas_subset = self.__client.map(get_meta_from_path, paths[index:index + 10000])
                ds_metas_futures += ds_metas_subset
                prog_bar.step()
            
            ds_metas = self.__client.gather(ds_metas_futures, direct=True)
            del ds_metas_futures
        
        for index, path in enumerate(paths):
            if ds_metas[index] is not None:
                self.__hf_to_meta_map[path] = ds_metas[index]
        self.__meta_pulled = True
        if check_valid:
            self.check_validity()
        else:
            logger.warning(f"Skipping metadata validation may result in errors due to missing attributes or coordinate data.")
        logger.info(f"Metadata pulled.")

    def check_validity(self):
        """Checks validity of metadata for each history file. Removes missing or incomplete metadata."""
        logger.debug(f"Validating metadata...")
        new_map = {}
        removed = {}
        for path in self.__hf_to_meta_map:
            if self.__hf_to_meta_map[path] is not None and self.__hf_to_meta_map[path].is_valid():
                new_map[path] = self.__hf_to_meta_map[path]
            else:
                removed[path] = self.__hf_to_meta_map[path]
                logger.log(LOG_LEVEL_IO_WARNING, f"Could not pull valid/complete metadata for '{path}'.")
        self.__hf_to_meta_map = new_map
        logger.debug(f"{len(new_map)} files valdiated ({len(removed)} removed).")
        return removed
    
    def include_patterns(self, glob_patterns):
        warnings.warn("TSCollection.include_patterns is deprecated in favor of TSCollection.include")
        return self.include(glob_patterns)

    def exclude_patterns(self, glob_patterns):
        warnings.warn("TSCollection.exclude_patterns is deprecated in favor of TSCollection.exclude")
        return self.exclude(glob_patterns)

    def include(self, glob_patterns):
        """
        Filters out history files in the collection with paths that do not match the glob patterns.

        :param glob_patterns: List of patterns to compare paths against.
        """
        filtered_path_map = {}
        for path in self.__hf_to_meta_map:
            for pattern in glob_patterns:
                if fnmatch.fnmatch(str(path), pattern):
                    filtered_path_map[path] = self.__hf_to_meta_map[path]
        logger.debug(f"Inclusive filter(s) applied: '{glob_patterns}'")
        return self.copy(meta_map=filtered_path_map)

    def exclude(self, glob_patterns):
        """
        Filters out history files in the collection with paths that do match the glob patterns.

        :param glob_patterns: List of patterns to compare paths against.
        """
        filtered_path_map = {}
        for path in self.__hf_to_meta_map:
            for pattern in glob_patterns:
                if not fnmatch.fnmatch(str(path), pattern):
                    filtered_path_map[path] = self.__hf_to_meta_map[path]
        logger.debug(f"Exclusive filter(s) applied: '{glob_patterns}'")
        return self.copy(meta_map=filtered_path_map)

    def include_years(self, start_year, end_year, glob_patterns=["*"]):
        """
        Filters out history files in the collection that fall outside the specified range of years.
        Glob patterns can be used to limit this filter to specific history files.
        

        :param start_year: First year in range.
        :param end_year: Last year in range.
        :param glob_patterns: Glob patterns to match history files that recieve this filter.
        """
        self.check_pulled()
        filtered_path_map = self.__hf_to_meta_map
        remove_paths = []
        for pattern in glob_patterns:
            for path in filtered_path_map:
                if fnmatch.fnmatch(path, pattern):
                    meta_ds = filtered_path_map[path]
                    if meta_ds.get_cftime_bounds() is not None:
                        avg_year = np.mean([ts.year for ts in meta_ds.get_cftime_bounds()[0]])
                    else:
                        avg_year = np.mean([ts.year for ts in meta_ds.get_cftimes()])
                    
                    if not start_year <= avg_year <= end_year:
                        remove_paths.append(path)

        for path in remove_paths:
            del filtered_path_map[path]

        logger.debug(f"Filtered from {start_year} to {end_year} applied to following glob patterns: '{glob_patterns}'")
        return self.copy(meta_map=filtered_path_map)

    def get_groups(self, check_fragmented=True):
        """
        Returns history file groupings.

        :param check_fragmented: Checks if any of the history files are spatially fragmented/tiled.
        :return: Dictionary containing group ID (key) and history file metadatas (value).
        """
        if self.__hf_groups is None:
            self.check_pulled()
            self.__hf_groups = sort_hf_groups(list(self.__hf_to_meta_map.keys()))
        
        if check_fragmented:
            self.check_pulled()
            self.__hf_groups = merge_fragmented_groups(self.__hf_groups, self.__hf_to_meta_map)

        return self.__hf_groups

    def slice_groups(self, slice_size_years=10, start_year=0, pattern=None):
        """
        Slices history file groupings that match the glob pattern (if specified) into subsets by time.

        :param slice_size_years: Size of slices to make, in years.
        :param start_year: Year to start slicing pattern from. Default is 0.
        :param pattern: Glob pattern to match history file grouping IDs to.
        :return: New groupings that are subset into time periods specified by 'slice_size_years'
        """
        sliced_groups = {}
        self.check_pulled()

        for group in self.get_groups():
            hf_paths = self.get_groups()[group]
            if pattern is not None and not fnmatch.fnmatch(group, pattern):
                sliced_groups[group] = hf_paths
                continue
            
            if len(hf_paths) == 1:
                sliced_groups[group] = hf_paths
                warnings.warn("Cannot slice history file group of size 1.", RuntimeWarning)
                continue
            else:
                group_meta_map = {path: self.__hf_to_meta_map[path] for path in hf_paths}
                
                min_year, max_year = get_year_bounds(group_meta_map)
                if start_year is not None:
                    min_year = start_year
                
                time_slices = calculate_year_slices(slice_size_years, min_year, max_year)

                hf_slices = {}
                variable_set = None
                for hf_path in hf_paths:
                    meta_ds = self.__hf_to_meta_map[hf_path]
        
                    if meta_ds.get_cftime_bounds() is not None:
                        time_bnds = meta_ds.get_cftime_bounds()[0]
                        time = time_bnds[0] + ((time_bnds[1] - time_bnds[0]) / 2)
                    else:
                        time = meta_ds.get_cftimes()[0]
                    
                    for time_slice in time_slices:
                        if time_slice[0] <= time.year <= time_slice[1]:
                            if time_slice in hf_slices:
                                hf_slices[time_slice].append(hf_path)
                            else:
                                hf_slices[time_slice] = [hf_path]
                            break

                for time_slice in hf_slices:
                    sliced_groups[f"{group}[sorting_pivot]{time_slice[0]}-{time_slice[1]}"] = hf_slices[time_slice]
        logger.debug(f"Slicing groups into {slice_size_years} year long slices for '{pattern}'.")
        return self.copy(hf_groups=sliced_groups)



from gents.tests.test_cases import *
from gents.hfcollection import HFCollection, find_files
from gents.timeseries import *
from os.path import isfile, getsize, isdir
from os import listdir, remove, makedirs
from netCDF4 import Dataset
from shutil import rmtree


def clear_output_dir(output_dir):
    """Helper function to clear the output directory after testing"""
    for name in listdir(output_dir):
        if isfile(f"{output_dir}/{name}"):
            remove(f"{output_dir}/{name}")
        else:
            rmtree(f"{output_dir}/{name}")


def test_clear_output_dir(tmp_path_factory):
    output_dir = tmp_path_factory.mktemp("output")
    for i in range(5):
        makedirs(f"{output_dir}/{i}/")
        for j in range(10):
            with open(f"{output_dir}/{i}/{j}.nc", 'w') as f:
                f.write("")
    
    assert len(find_files(output_dir, "*")) == 5*10
    clear_output_dir(output_dir)
    assert len(find_files(output_dir, "*")) == 0


def test_generate_time_series(simple_case):
    """This does not test TSCollection, but the primary function it relies on."""
    input_head_dir, output_head_dir = simple_case
    hf_paths = [f"{input_head_dir}/{name}" for name in listdir(input_head_dir)]

    var_name = "VAR1"
    ts_path = generate_time_series(hf_paths, f"{output_head_dir}/test_ts.", var_name, ["time", "time_bounds"], complevel=0, compression=None, overwrite=False)
    
    assert isfile(ts_path)
    assert check_timeseries_integrity(ts_path)

    with Dataset(ts_path, 'r') as ts_ds:
        assert ts_ds["time"][:].size == len(hf_paths)
        
        for index in range(len(hf_paths)):
            with Dataset(hf_paths[index], 'r') as hf_ds:
                assert (ts_ds[var_name][:][index] == hf_ds[var_name][:]).all()
        assert "time" in ts_ds.variables
        assert "time_bounds" in ts_ds.variables

    original_size = getsize(ts_path)
    ts_path = generate_time_series(hf_paths, f"{output_head_dir}/test_ts.", var_name, ["time", "time_bounds"], complevel=9, compression="zlib", overwrite=True)

    assert getsize(ts_path) < original_size
    assert len(listdir(output_head_dir)) == 1
    assert check_timeseries_integrity(ts_path)


def test_integrity_check(simple_case):
    input_head_dir, output_head_dir = simple_case
    hf_paths = [f"{input_head_dir}/{name}" for name in listdir(input_head_dir)]

    ts_path = generate_time_series(hf_paths, f"{output_head_dir}/test_ts.", "VAR1", ["time", "time_bounds"])
    assert check_timeseries_integrity(ts_path)
    for path in hf_paths:
        assert not check_timeseries_integrity(path)


def test_tscollection_copy(simple_case):
    input_head_dir, output_head_dir = simple_case
    hf_collection = HFCollection(input_head_dir)
    ts_collection = TSCollection(hf_collection, output_head_dir)

    ts_copy = ts_collection.copy()
    assert type(ts_copy) == TSCollection
    assert list(ts_copy) == list(ts_collection)
    assert ts_copy is not ts_collection

    ts_copy = ts_collection.include("*.nc", "*")
    assert type(ts_copy) == TSCollection
    assert list(ts_copy) == list(ts_collection)
    assert ts_copy is not ts_collection

    ts_copy = ts_collection.exclude("*.txt", "txt")
    assert type(ts_copy) == TSCollection
    assert list(ts_copy) == list(ts_collection)
    assert ts_copy is not ts_collection
    
    ts_copy = ts_collection.apply_compression(1, "zlib", "*.txt", "txt")
    assert type(ts_copy) == TSCollection
    assert list(ts_copy) == list(ts_collection)
    assert ts_copy is not ts_collection
    
    ts_copy = ts_collection.apply_overwrite("*.txt", "txt")
    assert type(ts_copy) == TSCollection
    assert list(ts_copy) == list(ts_collection)
    assert ts_copy is not ts_collection

    ts_copy = ts_collection.remove_overwrite("*.txt", "txt")
    assert type(ts_copy) == TSCollection
    assert list(ts_copy) == list(ts_collection)
    assert ts_copy is not ts_collection

    ts_copy = ts_collection.apply_path_swap("a", "b", "*.txt", "txt")
    assert type(ts_copy) == TSCollection
    assert list(ts_copy) == list(ts_collection)
    assert ts_copy is not ts_collection
    

def test_tscollection_compression(simple_case):
    """Assumes default compression is 0."""
    input_head_dir, output_head_dir = simple_case
    hf_collection = HFCollection(input_head_dir)
    ts_collection = TSCollection(hf_collection, output_head_dir)

    ts_collection.execute()

    uncompressed_size = 0
    for file_name in listdir(output_head_dir):
        uncompressed_size += getsize(f"{output_head_dir}/{file_name}")

    clear_output_dir(output_head_dir)

    compressed_collection = ts_collection.apply_compression(9, "zlib", "*", "*")
    compressed_collection.execute()

    compressed_size = 0
    for file_name in listdir(output_head_dir):
        compressed_size += getsize(f"{output_head_dir}/{file_name}")

    assert compressed_size > 0
    assert compressed_size < uncompressed_size


def test_tscollection_overwrite(simple_case):
    """
    To test the overwrite function, we write uncompressed and then overwrite with compressed.

    Since this entangles overwrite with compression, check if `test_tscollection_compression`
    passes, as it purely tests compression. 
    """
    input_head_dir, output_head_dir = simple_case
    hf_collection = HFCollection(input_head_dir)
    ts_collection = TSCollection(hf_collection, output_head_dir)

    ts_collection.execute()
    uncompressed_size = 0
    for file_name in listdir(output_head_dir):
        uncompressed_size += getsize(f"{output_head_dir}/{file_name}")

    compressed_collection = ts_collection.apply_compression(9, "zlib", "*", "*").apply_overwrite("*", "*")
    compressed_collection.execute()
    compressed_size = 0
    for file_name in listdir(output_head_dir):
        compressed_size += getsize(f"{output_head_dir}/{file_name}")

    assert compressed_size > 0
    assert compressed_size < uncompressed_size


def test_tscollection_filters(structured_case):
    input_head_dir, output_head_dir = structured_case
    hf_collection = HFCollection(input_head_dir)
    ts_collection = TSCollection(hf_collection, output_head_dir)

    ts_collection.execute()
    unfiltered_num_files = len(find_files(output_head_dir, "*"))
    clear_output_dir(output_head_dir)

    ts_collection.exclude("*/1_dir/*").execute()
    exclude_num_files = len(find_files(output_head_dir, "*"))
    clear_output_dir(output_head_dir)

    ts_collection.include("*/1_dir/*").execute()
    include_num_files = len(find_files(output_head_dir, "*"))

    assert unfiltered_num_files > exclude_num_files
    assert exclude_num_files > include_num_files
    assert include_num_files > 0


def test_ts_collection_path_swapping(structured_case):
    input_head_dir, output_head_dir = structured_case
    hf_collection = HFCollection(input_head_dir)
    ts_collection = TSCollection(hf_collection, output_head_dir)

    ts_collection.apply_path_swap("/0_subdir/", "/proc/tseries/").execute()
    assert isdir(f"{output_head_dir}/0_dir/proc/tseries/")
    assert len(listdir(f"{output_head_dir}/0_dir/proc/tseries/")) == STRUCTURED_NUM_TEST_HIST_FILES

    clear_output_dir(output_head_dir)

    ts_collection.apply_path_swap("dir", "folder").execute()
    assert isdir(f"{output_head_dir}/0_folder/0_subfolder/")
    assert len(listdir(output_head_dir)) == STRUCTURED_NUM_DIRS
    for top_dir in listdir(output_head_dir):
        assert len(listdir(f"{output_head_dir}/{top_dir}")) == STRUCTURED_NUM_SUBDIRS
        for sub_dir in listdir(f"{output_head_dir}/{top_dir}/"):
            assert len(listdir(f"{output_head_dir}/{top_dir}/{sub_dir}")) == STRUCTURED_NUM_VARS


def test_ts_collection_append_timestep_dirs(mixed_timestep_case):
    input_head_dir, output_head_dir = mixed_timestep_case
    hf_collection = HFCollection(input_head_dir)
    ts_collection = TSCollection(hf_collection, output_head_dir)

    ts_collection.append_timestep_dirs().execute()

    assert isdir(f"{output_head_dir}/hour_1")
    assert isdir(f"{output_head_dir}/day_1")
    assert isdir(f"{output_head_dir}/month_1")
    assert isdir(f"{output_head_dir}/year_1")

    assert len(listdir(f"{output_head_dir}/hour_1")) == SIMPLE_NUM_VARS
    assert len(listdir(f"{output_head_dir}/day_1")) == SIMPLE_NUM_VARS
    assert len(listdir(f"{output_head_dir}/month_1")) == SIMPLE_NUM_VARS
    assert len(listdir(f"{output_head_dir}/year_1")) == SIMPLE_NUM_VARS



from gents.utils import get_version
from gents.hfcollection import HFCollection
from gents.timeseries import TSCollection
from gents.tests.test_cases import *
from netCDF4 import Dataset
from os import listdir, makedirs, rename
import pytest
import numpy as np
import random


# Need to add tests for netCDF_3, netCDF4, netCDF4_classic on agg_dim for mfdataset

def is_monotonic(series):
    return (np.diff(series) > 0).all()


def test_monotonic_check():
    assert is_monotonic(np.arange(10))
    assert is_monotonic([-5, -2, 0, 5, 100])
    assert not is_monotonic([0, 0, 0])
    assert not is_monotonic([5, 2, 0, -5, -100])
    assert not is_monotonic([1, -1, 1, -1, 1])


def test_simple_workflow(simple_case):
    input_head_dir, output_head_dir = simple_case
    hf_collection = HFCollection(input_head_dir)
    ts_collection = TSCollection(hf_collection, output_head_dir)
    ts_paths = ts_collection.execute()

    hf_collection.sort_along_time()

    assert len(ts_paths) == SIMPLE_NUM_VARS
    
    for path in ts_paths:
        assert "*" not in path
        with Dataset(path, 'r') as ts_ds:
            assert ts_ds["time"].size == SIMPLE_NUM_TEST_HIST_FILES
            assert ts_ds["time_bounds"].shape[0] == SIMPLE_NUM_TEST_HIST_FILES
            assert ts_ds.getncattr("gents_version") == get_version()
            var_name = path.split(".")[-3]
            
            for index in range(ts_ds["time"].size):
                time_bounds = ts_ds["time_bounds"][index]
                assert time_bounds.count() == 2
                assert time_bounds[0] <= ts_ds["time"][index] <= time_bounds[1]
                
                with Dataset(list(hf_collection)[index], 'r') as hf_ds:
                    assert (ts_ds[var_name][index] == hf_ds[var_name]).all()

                    for key in hf_ds.ncattrs():
                        assert ts_ds.getncattr(key) == hf_ds.getncattr(key)
                    
                    for key in hf_ds[var_name].ncattrs():
                        assert ts_ds[var_name].getncattr(key) == hf_ds[var_name].getncattr(key)


def test_simple_workflow_slicing(simple_case):
    input_head_dir, output_head_dir = simple_case
    hf_collection = HFCollection(input_head_dir).slice_groups(slice_size_years=1)
    ts_collection = TSCollection(hf_collection, output_head_dir)
    ts_paths = ts_collection.execute()

    assert len(ts_paths) == SIMPLE_NUM_VARS*int(np.ceil(SIMPLE_NUM_TEST_HIST_FILES / 12))
    for path in ts_paths:
        assert "[sorting_pivot]" not in path
        assert "*" not in path
        with Dataset(path, 'r') as ts_ds:
            assert ts_ds["time"].size == ts_ds["time_bounds"].shape[0]

            for index in range(ts_ds["time"].size):
                time_bounds = ts_ds["time_bounds"][index]
                assert time_bounds.count() == 2
                assert time_bounds[0] <= ts_ds["time"][index] <= time_bounds[1]


def test_unstructured_grid_workflow(unstructured_grid_case):
    input_head_dir, output_head_dir = unstructured_grid_case

    hf_collection = HFCollection(input_head_dir)
    ts_collection = TSCollection(hf_collection, output_head_dir)
    ts_paths = ts_collection.execute()

    assert len(ts_paths) == SIMPLE_NUM_VARS


def test_time_bounds_workflow(time_bounds_case):
    input_head_dir, output_head_dir = time_bounds_case
    hf_collection = HFCollection(input_head_dir)
    ts_collection = TSCollection(hf_collection, output_head_dir)
    ts_paths = ts_collection.execute()

    for path in ts_paths:
        with Dataset(path, 'r') as ts_ds:
            assert "Time_Bounds" in ts_ds.variables
            assert "Time" in ts_ds.variables
            assert "Time" in ts_ds.dimensions


def test_no_time_bounds_workflow(no_time_bounds_case):
    input_head_dir, output_head_dir = no_time_bounds_case
    hf_collection = HFCollection(input_head_dir)
    ts_collection = TSCollection(hf_collection, output_head_dir)
    ts_paths = ts_collection.execute()

    hf_collection.sort_along_time()
    hf_collection = hf_collection.include_years(0, 99999)
    hf_collection = hf_collection.slice_groups(99999)

    assert len(ts_paths) == SIMPLE_NUM_VARS
    
    for path in ts_paths:
        with Dataset(path, 'r') as ts_ds:
            assert ts_ds["time"].size == SIMPLE_NUM_TEST_HIST_FILES
            assert ts_ds.getncattr("gents_version") == get_version()
            var_name = path.split(".")[-3]
            
            for index in range(ts_ds["time"].size):
                with Dataset(list(hf_collection)[index], 'r') as hf_ds:
                    assert (ts_ds[var_name][index] == hf_ds[var_name]).all()

                    for key in hf_ds.ncattrs():
                        assert ts_ds.getncattr(key) == hf_ds.getncattr(key)
                    
                    for key in hf_ds[var_name].ncattrs():
                        assert ts_ds[var_name].getncattr(key) == hf_ds[var_name].getncattr(key)


def test_scrambled_workflow(scrambled_case):
    input_head_dir, output_head_dir = scrambled_case
    hf_collection = HFCollection(input_head_dir)
    ts_collection = TSCollection(hf_collection, output_head_dir)
    ts_paths = ts_collection.execute()

    hf_collection.sort_along_time()

    assert len(ts_paths) == SCRAMBLED_NUM_VARS

    for path in ts_paths:
        with Dataset(path, 'r') as ts_ds:
            assert ts_ds["time"].size == SCRAMBLED_NUM_TEST_HIST_FILES
            assert is_monotonic(ts_ds["time"][:])


def test_structured_workflow(structured_case):
    input_head_dir, output_head_dir = structured_case
    hf_collection = HFCollection(input_head_dir)
    ts_collection = TSCollection(hf_collection, output_head_dir)
    ts_paths = ts_collection.execute()
    
    assert len(ts_paths) == SCRAMBLED_NUM_VARS*STRUCTURED_NUM_DIRS*STRUCTURED_NUM_SUBDIRS


def test_multistep_workflow(multistep_case):
    input_head_dir, output_head_dir = multistep_case
    hf_collection = HFCollection(input_head_dir)
    ts_collection = TSCollection(hf_collection, output_head_dir)
    ts_paths = ts_collection.execute()

    hf_collection.sort_along_time()

    assert len(ts_paths) == SIMPLE_NUM_VARS

    for path in ts_paths:
        with Dataset(path, 'r') as ts_ds:
            assert is_monotonic(ts_ds["time"][:])


def test_with_auxiliary_workflow(with_auxiliary_case):
    input_head_dir, output_head_dir = with_auxiliary_case
    hf_collection = HFCollection(input_head_dir)
    ts_collection = TSCollection(hf_collection, output_head_dir)
    ts_paths = ts_collection.execute()

    assert len(ts_paths) == 1

    with Dataset(ts_paths[0], 'r') as ts_ds:
        assert is_monotonic(ts_ds["time"][:])
        for var_index in range(SIMPLE_NUM_VARS):
            assert f"VAR_AUX_{var_index}" in ts_ds.variables
            assert len(ts_ds[f"VAR_AUX_{var_index}"].shape) == 1


def test_modified_extensions_workflow(simple_case):
    input_head_dir, output_head_dir = simple_case

    for index, file_name in enumerate(listdir(input_head_dir)):
        rename(f"{input_head_dir}/{file_name}", f"{input_head_dir}/test.hf.modified_ext_simple.nc.{index}")

    hf_collection = HFCollection(input_head_dir)
    ts_collection = TSCollection(hf_collection, output_head_dir)
    ts_paths = ts_collection.execute()

    assert len(ts_paths) == SIMPLE_NUM_VARS
    
    for path in ts_paths:
        with Dataset(path, 'r') as ts_ds:
            assert ts_ds["time"].size == SIMPLE_NUM_TEST_HIST_FILES


def test_spatially_fragmented_workflow(spatial_fragment_case):
    input_head_dir, output_head_dir = spatial_fragment_case
    hf_collection = HFCollection(input_head_dir)
    
    hf_collection.get_groups(check_fragmented=True)
    ts_collection = TSCollection(hf_collection, output_head_dir)
    ts_paths = ts_collection.execute()

    assert len(ts_paths) == SIMPLE_NUM_VARS

    for path in ts_paths:
        with Dataset(path, 'r') as ds:
            assert ds["lat"].size == FRAGMENTED_NUM_LAT_FILES*FRAGMENTED_NUM_LAT_PTS_PER_HF
            assert ds["lon"].size == FRAGMENTED_NUM_LON_FILES*FRAGMENTED_NUM_LON_PTS_PER_HF
            assert ds["time"].size == FRAGMENTED_NUM_TIMESTEPS


def test_auxiliary_only_workflow(auxiliary_only_case):
    input_head_dir, output_head_dir = auxiliary_only_case
    hf_collection = HFCollection(input_head_dir)
    ts_collection = TSCollection(hf_collection, output_head_dir)
    ts_collection.include("*").exclude("").apply_overwrite("*").execute()

    assert len(listdir(output_head_dir)) == 1
    assert "auxiliary" in listdir(output_head_dir)[0]


from netCDF4 import Dataset
from os import makedirs
import numpy as np
import pytest
import random

CASE_START_YEAR = 1850
SIMPLE_NUM_TEST_HIST_FILES = 49
SIMPLE_NUM_VARS = 6
TIME_NUM_TEST_HIST_FILES = 2
TIME_NUM_VARS = 1
SCRAMBLED_NUM_TEST_HIST_FILES = 36
SCRAMBLED_NUM_VARS = 2
STRUCTURED_NUM_TEST_HIST_FILES = 2
STRUCTURED_NUM_VARS = 2
STRUCTURED_NUM_DIRS = 3
STRUCTURED_NUM_SUBDIRS = 2
FRAGMENTED_NUM_LAT_FILES = 3
FRAGMENTED_NUM_LON_FILES = 2
FRAGMENTED_NUM_LAT_PTS_PER_HF = 2
FRAGMENTED_NUM_LON_PTS_PER_HF = 1
FRAGMENTED_NUM_TIMESTEPS = 20
UNSTRUCT_GRID_NUM_NCOLS = 8
MIXED_TS_NUM_TEST_HIST_FILES = 10


def generate_history_file(
        path,
        time_val,
        time_bounds_val,
        num_vars=SIMPLE_NUM_VARS,
        nc_format="NETCDF4_CLASSIC",
        time_bounds_attrs=True,
        time_name="time",
        time_bounds_name="time_bounds",
        auxiliary=False,
        aux_dim="time",
        dim_shapes=None,
        dim_vals={},
        var_dims=None,
        var_shape=None,
        disable_primary_var=False
    ):
    if dim_shapes is None: 
        dim_shapes = {
            time_name: None,
            "bnds": 2,
            "lat": 3,
            "lon": 4,
            "lev": 1
        }
    
    with Dataset(path, "w", format=nc_format) as ds:
        for dim in dim_shapes:
            ds.createDimension(dim, dim_shapes[dim])
            if dim in dim_vals:
                dim_data = ds.createVariable(dim, float, (dim))
                dim_data[:] = dim_vals[dim]

        if var_shape is None:
            var_shape = (len(time_val), dim_shapes["lat"], dim_shapes["lon"])

        if var_dims is None:
            var_dims = (time_name, "lat", "lon")

        for index in range(num_vars):
            if auxiliary:
                var_data = ds.createVariable(f"VAR_AUX_{index}", float, (aux_dim))
                aux_shape = 1
                if dim_shapes[aux_dim] is not None:
                    aux_shape = dim_shapes[aux_dim]
                var_data[:] = np.random.random((aux_shape)).astype(float)
                var_data.setncatts({
                    "units": "kg/g/m^2/K",
                    "standard_name": f"VAR{index}",
                    "long_name": f"variable_{index}"
                })
            if not disable_primary_var:
                var_data = ds.createVariable(f"VAR{index}", float, var_dims)
                var_data[:] = index*np.ones(var_shape).astype(float)
                var_data.setncatts({
                    "units": "kg/g/m^2/K",
                    "standard_name": f"VAR{index}",
                    "long_name": f"variable_{index}"
                })

        if time_val is not None:
            time_data = ds.createVariable(time_name, np.double, time_name)
            time_data[:] = time_val
            time_data.setncatts({
                "calendar": "360_day",
                "units": f"days since {CASE_START_YEAR}-01-01",
                "standard_name": time_name,
                "long_name": time_name
            })

        if time_bounds_val is not None:
            time_bnds_data = ds.createVariable(time_bounds_name, np.double, (time_name, "bnds"))
            time_bnds_data[:] = time_bounds_val
            if time_bounds_attrs:
                time_bnds_data.setncatts({
                    "calendar": "360_day",
                    "units": "days since 1850-01-01",
                    "standard_name": time_bounds_name,
                    "long_name": time_bounds_name
                })
            
        ds.setncatts({
            "source": "GenTS testing suite",
            "description": "Synthetic data used for testing with the GenTS package."
        })


@pytest.fixture(scope="function")
def time_bounds_case(tmp_path_factory):
    head_hf_dir = tmp_path_factory.mktemp("time_history_files")
    head_ts_dir = tmp_path_factory.mktemp("time_timeseries_files")
    
    hf_paths = [f"{head_hf_dir}/testing.hf.{str(index).zfill(5)}.nc" for index in range(TIME_NUM_TEST_HIST_FILES)]
    for file_index, path in enumerate(hf_paths):
        generate_history_file(path, [(file_index+0.5)*30], [[file_index*30, (file_index+1)*30]], num_vars=TIME_NUM_VARS, time_bounds_name="Time_Bounds", time_name="Time")

    return head_hf_dir, head_ts_dir


@pytest.fixture(scope="function")
def simple_case(tmp_path_factory):
    head_hf_dir = tmp_path_factory.mktemp("simple_history_files")
    head_ts_dir = tmp_path_factory.mktemp("simple_timeseries_files")
    
    hf_paths = [f"{head_hf_dir}/testing.hf.{str(index).zfill(5)}.nc" for index in range(SIMPLE_NUM_TEST_HIST_FILES)]
    for file_index, path in enumerate(hf_paths):
        generate_history_file(path, [(file_index+0.5)*30], [[file_index*30, (file_index+1)*30]])

    return head_hf_dir, head_ts_dir


@pytest.fixture(scope="function")
def unstructured_grid_case(tmp_path_factory):
    head_hf_dir = tmp_path_factory.mktemp("unstructured_grid_history_files")
    head_ts_dir = tmp_path_factory.mktemp("unstructured_grid_timeseries_files")
    
    hf_paths = [f"{head_hf_dir}/testing.hf.{str(index).zfill(5)}.nc" for index in range(SIMPLE_NUM_TEST_HIST_FILES)]
    for file_index, path in enumerate(hf_paths):
        time_vals, time_bnds = [(file_index+0.5)*30], [[file_index*30, (file_index+1)*30]]
        generate_history_file(path, time_vals, time_bnds, auxiliary=True, aux_dim="ncol", dim_shapes={"time": None, "ncol": UNSTRUCT_GRID_NUM_NCOLS, "bnds": 2}, var_dims=("time", "ncol"), var_shape=(len(time_vals), UNSTRUCT_GRID_NUM_NCOLS))

    return head_hf_dir, head_ts_dir


@pytest.fixture(scope="function")
def scrambled_case(tmp_path_factory):
    head_hf_dir = tmp_path_factory.mktemp("scrambled_history_files")
    head_ts_dir = tmp_path_factory.mktemp("scrambled_timeseries_files")
    
    hf_paths = [f"{head_hf_dir}/testing.hf.{str(index).zfill(5)}.nc" for index in range(SCRAMBLED_NUM_TEST_HIST_FILES)]

    random.seed(0)
    random.shuffle(hf_paths)
    
    for file_index, path in enumerate(hf_paths):
        generate_history_file(path, [(file_index+0.5)*30], [[file_index*30, (file_index+1)*30]], num_vars=SCRAMBLED_NUM_VARS)
    return head_hf_dir, head_ts_dir
    

@pytest.fixture(scope="function")
def structured_case(tmp_path_factory):
    head_hf_dir = tmp_path_factory.mktemp("structured_history_files")
    head_ts_dir = tmp_path_factory.mktemp("structured_timeseries_files")

    for top_index in range(STRUCTURED_NUM_DIRS):
        for sub_index in range(STRUCTURED_NUM_SUBDIRS):
            dir_path = f"{head_hf_dir}/{top_index}_dir/{sub_index}_subdir/"
            makedirs(dir_path, exist_ok=True)
            
            for file_index in range(STRUCTURED_NUM_TEST_HIST_FILES):
                path = f"{dir_path}/testing.hf.{str(file_index).zfill(5)}.nc" 
                generate_history_file(path, [(file_index+0.5)*30], [[file_index*30, (file_index+1)*30]], num_vars=SCRAMBLED_NUM_VARS)
    return head_hf_dir, head_ts_dir


@pytest.fixture(scope="function")
def no_time_bounds_case(tmp_path_factory):
    head_hf_dir = tmp_path_factory.mktemp("no_tb_history_files")
    head_ts_dir = tmp_path_factory.mktemp("no_tb_timeseries_files")
    
    hf_paths = [f"{head_hf_dir}/testing.hf.{str(index).zfill(5)}.nc" for index in range(SIMPLE_NUM_TEST_HIST_FILES)]
    for file_index, path in enumerate(hf_paths):
        generate_history_file(path, [(file_index+0.5)*30], None)

    return head_hf_dir, head_ts_dir


@pytest.fixture(scope="function")
def no_time_case(tmp_path_factory):
    head_hf_dir = tmp_path_factory.mktemp("no_time_history_files")
    head_ts_dir = tmp_path_factory.mktemp("no_time_timeseries_files")
    
    hf_paths = [f"{head_hf_dir}/testing.hf.{str(index).zfill(5)}.nc" for index in range(SIMPLE_NUM_TEST_HIST_FILES)]
    for file_index, path in enumerate(hf_paths):
        generate_history_file(path, [(file_index+0.5)*30], None, time_name="nottime")

    return head_hf_dir, head_ts_dir


@pytest.fixture(scope="function")
def simple_case_missing_attrs(tmp_path_factory):
    head_hf_dir = tmp_path_factory.mktemp("simple_history_files")
    head_ts_dir = tmp_path_factory.mktemp("simple_timeseries_files")
    
    hf_paths = [f"{head_hf_dir}/testing.hf.{str(index).zfill(5)}.nc" for index in range(SIMPLE_NUM_TEST_HIST_FILES)]
    for file_index, path in enumerate(hf_paths):
        generate_history_file(path, [(file_index+0.5)*30], [[file_index*30, (file_index+1)*30]], time_bounds_attrs=False)

    return head_hf_dir, head_ts_dir


@pytest.fixture(scope="function")
def multistep_case(tmp_path_factory):
    head_hf_dir = tmp_path_factory.mktemp("multistep_history_files")
    head_ts_dir = tmp_path_factory.mktemp("multistep_timeseries_files")
    
    hf_paths = [f"{head_hf_dir}/testing.hf.{str(index).zfill(5)}.nc" for index in range(SIMPLE_NUM_TEST_HIST_FILES)]
    index = 0
    for  path in hf_paths:
        generate_history_file(path, [(index)*30, (index+1)*30, (index+2)*30], [[(index)*30, (index+1)*30], [(index+1)*30, (index+2)*30], [(index+2)*30, (index+3)*30]])
        index += 3

    return head_hf_dir, head_ts_dir


@pytest.fixture(scope="function")
def with_auxiliary_case(tmp_path_factory):
    head_hf_dir = tmp_path_factory.mktemp("with_auxiliary_history_files")
    head_ts_dir = tmp_path_factory.mktemp("with_auxiliary_timeseries_files")
    
    hf_paths = [f"{head_hf_dir}/testing.hf.{str(index).zfill(5)}.nc" for index in range(SIMPLE_NUM_TEST_HIST_FILES)]
    for file_index, path in enumerate(hf_paths):
        time_vals, time_bnds = [(file_index+0.5)*30], [[file_index*30, (file_index+1)*30]]
        generate_history_file(path, time_vals, None, auxiliary=True, time_name="time", dim_shapes={"time": None}, var_dims=("time"), var_shape=(len(time_vals)))

    return head_hf_dir, head_ts_dir


@pytest.fixture(scope="function")
def spatial_fragment_case(tmp_path_factory):
    head_hf_dir = tmp_path_factory.mktemp("fragmented_history_files")
    head_ts_dir = tmp_path_factory.mktemp("fragmented_timeseries_files")
    
    base_hf_paths = [f"{head_hf_dir}/testing.hf.{str(index).zfill(5)}.nc" for index in range(FRAGMENTED_NUM_TIMESTEPS)]

    lat_range = np.linspace(-90, 90, FRAGMENTED_NUM_LAT_FILES*FRAGMENTED_NUM_LAT_PTS_PER_HF)
    lon_range = np.linspace(-180, 180, FRAGMENTED_NUM_LON_FILES*FRAGMENTED_NUM_LON_PTS_PER_HF)

    dim_shapes = {
        "time": None,
        "bnds": 2,
        "lat": FRAGMENTED_NUM_LAT_PTS_PER_HF,
        "lon": FRAGMENTED_NUM_LON_PTS_PER_HF
    }

    for file_index, path in enumerate(base_hf_paths):
        tile_index = 0
        for lat_index in range(0, lat_range.size, FRAGMENTED_NUM_LAT_PTS_PER_HF):
            for lon_index in range(0, lon_range.size, FRAGMENTED_NUM_LON_PTS_PER_HF):
                dim_vals = {
                    "lat": lat_range[lat_index:lat_index+FRAGMENTED_NUM_LAT_PTS_PER_HF],
                    "lon": lon_range[lon_index:lon_index+FRAGMENTED_NUM_LON_PTS_PER_HF]
                }
                generate_history_file(f"{path}.{tile_index}", [(file_index+1)*180], [[file_index*180, (file_index+1)*180]], dim_shapes=dim_shapes, dim_vals=dim_vals)
                tile_index += 1

    return head_hf_dir, head_ts_dir


@pytest.fixture(scope="function")
def mixed_timestep_case(tmp_path_factory):
    head_hf_dir = tmp_path_factory.mktemp("mixed_timestep_history_files")
    head_ts_dir = tmp_path_factory.mktemp("mixed_timestep_timeseries_files")
    
    hf_paths = [f"{head_hf_dir}/testing.hf0.{str(index).zfill(5)}.nc" for index in range(MIXED_TS_NUM_TEST_HIST_FILES)]
    for file_index, path in enumerate(hf_paths):
        generate_history_file(path, [(file_index+1)*0.1], [[file_index*0.1, (file_index+1)*0.1]])

    hf_paths = [f"{head_hf_dir}/testing.hf1.{str(index).zfill(5)}.nc" for index in range(MIXED_TS_NUM_TEST_HIST_FILES)]
    for file_index, path in enumerate(hf_paths):
        generate_history_file(path, [(file_index+1)*1], [[file_index*1, (file_index+1)*1]])

    hf_paths = [f"{head_hf_dir}/testing.hf2.{str(index).zfill(5)}.nc" for index in range(MIXED_TS_NUM_TEST_HIST_FILES)]
    for file_index, path in enumerate(hf_paths):
        generate_history_file(path, [(file_index+0.5)*30], [[file_index*30, (file_index+1)*30]])

    hf_paths = [f"{head_hf_dir}/testing.hf3.{str(index).zfill(5)}.nc" for index in range(MIXED_TS_NUM_TEST_HIST_FILES)]
    for file_index, path in enumerate(hf_paths):
        generate_history_file(path, [(file_index+1)*365], [[file_index*365, (file_index+1)*365]])

    return head_hf_dir, head_ts_dir

@pytest.fixture(scope="function")
def auxiliary_only_case(tmp_path_factory):
    head_hf_dir = tmp_path_factory.mktemp("with_auxiliary_history_files")
    head_ts_dir = tmp_path_factory.mktemp("with_auxiliary_timeseries_files")
    
    hf_paths = [f"{head_hf_dir}/testing.hf.{str(index).zfill(5)}.nc" for index in range(SIMPLE_NUM_TEST_HIST_FILES)]
    for file_index, path in enumerate(hf_paths):
        generate_history_file(path, [(file_index+1)*1], None, var_dims=("time"), var_shape=1, auxiliary=True, aux_dim="time", disable_primary_var=True, dim_shapes={"time": None})

    return head_hf_dir, head_ts_dir


from gents.utils import *
import pytest

LOGGER_HEADER_NUM_LINES = 2


def test_timestamp():
    assert type(get_time_stamp()) == str


def test_version():
    assert type(get_version()) == str


@pytest.fixture(scope="session")
def log_output_dir(tmp_path_factory):
    return tmp_path_factory.mktemp("log_data")


def test_logging(log_output_dir):
    logger = logging.getLogger("gents")

    enable_logging(verbose=False, output_path=f"{log_output_dir}/log_header.txt")

    with open(f"{log_output_dir}/log_header.txt") as f:
        assert len(f.readlines()) == LOGGER_HEADER_NUM_LINES

    enable_logging(verbose=False, output_path=f"{log_output_dir}/non_verbose.txt")

    logger.debug("test")
    logger.info("test")
    logger.warning("test")
    logger.log(LOG_LEVEL_IO_WARNING, "test")

    with open(f"{log_output_dir}/non_verbose.txt") as f:
        assert len(f.readlines()) == LOGGER_HEADER_NUM_LINES + 3
    
    enable_logging(verbose=True, output_path=f"{log_output_dir}/verbose.txt")

    logger.debug("test")
    logger.info("test")
    logger.warning("test")
    logger.log(LOG_LEVEL_IO_WARNING, "test")

    with open(f"{log_output_dir}/verbose.txt") as f:
        assert len(f.readlines()) == LOGGER_HEADER_NUM_LINES + 4